"""

Choice model with a latent variable: maximum likelihood estimation
==================================================================

Mixture of logit.
Measurement equation for the indicators.
Maximum likelihood (full information) estimation.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 18:02:26 2023
"""

import sys
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
import biogeme.distributions as dist
import biogeme.results as res
from biogeme.expressions import (
    Beta,
    log,
    RandomVariable,
    Integrate,
    Elem,
    bioNormalCdf,
    exp,
)
from read_or_estimate import read_or_estimate

from optima import (
    database,
    age_65_more,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    WaitingTimePT,
    Envir01,
    Envir02,
    Envir03,
    Mobil11,
    Mobil14,
    Mobil16,
    Mobil17,
    Choice,
    TimePT_scaled,
    TimeCar_scaled,
    MarginalCostPT_scaled,
    CostCarCHF_scaled,
    distance_km_scaled,
    PurpHWH,
    PurpOther,
    ScaledIncome,
)


logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b05latent_choice_full.py')


# %%
# Read the estimates from the structural equation estimation.
MODELNAME = 'b02one_latent_ordered'
try:
    struct_results = res.bioResults(pickleFile=f'saved_results/{MODELNAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {MODELNAME}.py in order to generate the '
        f'file {MODELNAME}.pickle, and move it to the directory saved_results'
    )
    sys.exit()
struct_betas = struct_results.getBetaValues()

# %%
# Coefficients
coef_intercept = Beta('coef_intercept', struct_betas['coef_intercept'], None, None, 0)
coef_age_65_more = Beta(
    'coef_age_65_more', struct_betas['coef_age_65_more'], None, None, 0
)
coef_haveGA = Beta('coef_haveGA', struct_betas['coef_haveGA'], None, None, 0)

coef_moreThanOneCar = Beta(
    'coef_moreThanOneCar', struct_betas['coef_moreThanOneCar'], None, None, 0
)
coef_moreThanOneBike = Beta(
    'coef_moreThanOneBike', struct_betas['coef_moreThanOneBike'], None, None, 0
)
coef_individualHouse = Beta(
    'coef_individualHouse', struct_betas['coef_individualHouse'], None, None, 0
)
coef_male = Beta('coef_male', struct_betas['coef_male'], None, None, 0)
coef_haveChildren = Beta(
    'coef_haveChildren', struct_betas['coef_haveChildren'], None, None, 0
)
coef_highEducation = Beta(
    'coef_highEducation', struct_betas['coef_highEducation'], None, None, 0
)

# %%
# Latent variable: structural equation

# %%
# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

# %%
# Piecewise linear specification for income.
thresholds = [None, 4, 6, 8, 10, None]
betas_thresholds = [
    Beta(
        'beta_ScaledIncome_minus_inf_4',
        struct_betas['beta_ScaledIncome_minus_inf_4'],
        None,
        None,
        0,
    ),
    Beta(
        'beta_ScaledIncome_4_6',
        struct_betas['beta_ScaledIncome_4_6'],
        None,
        None,
        0,
    ),
    Beta(
        'beta_ScaledIncome_6_8',
        struct_betas['beta_ScaledIncome_6_8'],
        None,
        None,
        0,
    ),
    Beta(
        'beta_ScaledIncome_8_10',
        struct_betas['beta_ScaledIncome_8_10'],
        None,
        None,
        0,
    ),
    Beta(
        'beta_ScaledIncome_10_inf',
        struct_betas['beta_ScaledIncome_10_inf'],
        None,
        None,
        0,
    ),
]

formula_income = models.piecewiseFormula(
    variable=ScaledIncome,
    thresholds=thresholds,
    betas=betas_thresholds,
)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formula_income
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * omega
)

# %%
# Measurement equations.

# %%
# Intercepts.
INTER_Envir01 = Beta('INTER_Envir01', 0, None, None, 1)
INTER_Envir02 = Beta('INTER_Envir02', 0, None, None, 0)
INTER_Envir03 = Beta('INTER_Envir03', 0, None, None, 0)
INTER_Mobil11 = Beta('INTER_Mobil11', 0, None, None, 0)
INTER_Mobil14 = Beta('INTER_Mobil14', 0, None, None, 0)
INTER_Mobil16 = Beta('INTER_Mobil16', 0, None, None, 0)
INTER_Mobil17 = Beta('INTER_Mobil17', 0, None, None, 0)

# %%
# Coefficients.
B_Envir01_F1 = Beta('B_Envir01_F1', -1, None, None, 1)
B_Envir02_F1 = Beta('B_Envir02_F1', -1, None, None, 0)
B_Envir03_F1 = Beta('B_Envir03_F1', 1, None, None, 0)
B_Mobil11_F1 = Beta('B_Mobil11_F1', 1, None, None, 0)
B_Mobil14_F1 = Beta('B_Mobil14_F1', 1, None, None, 0)
B_Mobil16_F1 = Beta('B_Mobil16_F1', 1, None, None, 0)
B_Mobil17_F1 = Beta('B_Mobil17_F1', 1, None, None, 0)

# %%
# Linear models.
MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

# %%
# Scale parameters
SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01', 1, 1.0e-5, None, 1)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02', 1, 1.0e-5, None, 0)
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16', 1, 1.0e-5, None, 0)
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17', 1, 1.0e-5, None, 0)

# %%
# Symmetric thresholds.
delta_1 = Beta('delta_1', 0.1, 1.0e-5, None, 0)
delta_2 = Beta('delta_2', 0.2, 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

# %%
# Ordered probit models.
Envir01_tau_1 = (tau_1 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_2 = (tau_2 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_3 = (tau_3 - MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_4 = (tau_4 - MODEL_Envir01) / SIGMA_STAR_Envir01
IndEnvir01 = {
    1: bioNormalCdf(Envir01_tau_1),
    2: bioNormalCdf(Envir01_tau_2) - bioNormalCdf(Envir01_tau_1),
    3: bioNormalCdf(Envir01_tau_3) - bioNormalCdf(Envir01_tau_2),
    4: bioNormalCdf(Envir01_tau_4) - bioNormalCdf(Envir01_tau_3),
    5: 1 - bioNormalCdf(Envir01_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir01 = Elem(IndEnvir01, Envir01)

Envir02_tau_1 = (tau_1 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_2 = (tau_2 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_3 = (tau_3 - MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_4 = (tau_4 - MODEL_Envir02) / SIGMA_STAR_Envir02
IndEnvir02 = {
    1: bioNormalCdf(Envir02_tau_1),
    2: bioNormalCdf(Envir02_tau_2) - bioNormalCdf(Envir02_tau_1),
    3: bioNormalCdf(Envir02_tau_3) - bioNormalCdf(Envir02_tau_2),
    4: bioNormalCdf(Envir02_tau_4) - bioNormalCdf(Envir02_tau_3),
    5: 1 - bioNormalCdf(Envir02_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir02 = Elem(IndEnvir02, Envir02)

Envir03_tau_1 = (tau_1 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_2 = (tau_2 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_3 = (tau_3 - MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_4 = (tau_4 - MODEL_Envir03) / SIGMA_STAR_Envir03
IndEnvir03 = {
    1: bioNormalCdf(Envir03_tau_1),
    2: bioNormalCdf(Envir03_tau_2) - bioNormalCdf(Envir03_tau_1),
    3: bioNormalCdf(Envir03_tau_3) - bioNormalCdf(Envir03_tau_2),
    4: bioNormalCdf(Envir03_tau_4) - bioNormalCdf(Envir03_tau_3),
    5: 1 - bioNormalCdf(Envir03_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Envir03 = Elem(IndEnvir03, Envir03)

Mobil11_tau_1 = (tau_1 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_2 = (tau_2 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_3 = (tau_3 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_4 = (tau_4 - MODEL_Mobil11) / SIGMA_STAR_Mobil11
IndMobil11 = {
    1: bioNormalCdf(Mobil11_tau_1),
    2: bioNormalCdf(Mobil11_tau_2) - bioNormalCdf(Mobil11_tau_1),
    3: bioNormalCdf(Mobil11_tau_3) - bioNormalCdf(Mobil11_tau_2),
    4: bioNormalCdf(Mobil11_tau_4) - bioNormalCdf(Mobil11_tau_3),
    5: 1 - bioNormalCdf(Mobil11_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil11 = Elem(IndMobil11, Mobil11)

Mobil14_tau_1 = (tau_1 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_2 = (tau_2 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_3 = (tau_3 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_4 = (tau_4 - MODEL_Mobil14) / SIGMA_STAR_Mobil14
IndMobil14 = {
    1: bioNormalCdf(Mobil14_tau_1),
    2: bioNormalCdf(Mobil14_tau_2) - bioNormalCdf(Mobil14_tau_1),
    3: bioNormalCdf(Mobil14_tau_3) - bioNormalCdf(Mobil14_tau_2),
    4: bioNormalCdf(Mobil14_tau_4) - bioNormalCdf(Mobil14_tau_3),
    5: 1 - bioNormalCdf(Mobil14_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil14 = Elem(IndMobil14, Mobil14)

Mobil16_tau_1 = (tau_1 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_2 = (tau_2 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_3 = (tau_3 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_4 = (tau_4 - MODEL_Mobil16) / SIGMA_STAR_Mobil16
IndMobil16 = {
    1: bioNormalCdf(Mobil16_tau_1),
    2: bioNormalCdf(Mobil16_tau_2) - bioNormalCdf(Mobil16_tau_1),
    3: bioNormalCdf(Mobil16_tau_3) - bioNormalCdf(Mobil16_tau_2),
    4: bioNormalCdf(Mobil16_tau_4) - bioNormalCdf(Mobil16_tau_3),
    5: 1 - bioNormalCdf(Mobil16_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil16 = Elem(IndMobil16, Mobil16)

Mobil17_tau_1 = (tau_1 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_2 = (tau_2 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_3 = (tau_3 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_4 = (tau_4 - MODEL_Mobil17) / SIGMA_STAR_Mobil17
IndMobil17 = {
    1: bioNormalCdf(Mobil17_tau_1),
    2: bioNormalCdf(Mobil17_tau_2) - bioNormalCdf(Mobil17_tau_1),
    3: bioNormalCdf(Mobil17_tau_3) - bioNormalCdf(Mobil17_tau_2),
    4: bioNormalCdf(Mobil17_tau_4) - bioNormalCdf(Mobil17_tau_3),
    5: 1 - bioNormalCdf(Mobil17_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0,
}

P_Mobil17 = Elem(IndMobil17, Mobil17)

# %%
# Choice model
# Read the estimates from the sequential estimation, and use
# them as starting values
MODELNAME = 'b04latent_choice_seq'
try:
    choice_results = res.bioResults(pickleFile=f'saved_results/{MODELNAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {MODELNAME}.py in order to generate the '
        f'file {MODELNAME}.pickle, and move it to the directory saved_results'
    )
    sys.exit()
choice_betas = choice_results.getBetaValues()

# %%
# Parameters to estimate. We use the previously estimated values as starting points.
ASC_CAR = Beta('ASC_CAR', choice_betas['ASC_CAR'], None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', choice_betas['ASC_SM'], None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', choice_betas['BETA_COST_HWH'], None, None, 0)
BETA_COST_OTHER = Beta(
    'BETA_COST_OTHER', choice_betas['BETA_COST_OTHER'], None, None, 0
)
BETA_DIST = Beta('BETA_DIST', choice_betas['BETA_DIST'], None, None, 0)
BETA_TIME_CAR_REF = Beta(
    'BETA_TIME_CAR_REF', choice_betas['BETA_TIME_CAR_REF'], None, 0, 0
)
BETA_TIME_CAR_CL = Beta(
    'BETA_TIME_CAR_CL', choice_betas['BETA_TIME_CAR_CL'], None, None, 0
)
BETA_TIME_PT_REF = Beta(
    'BETA_TIME_PT_REF', choice_betas['BETA_TIME_PT_REF'], None, 0, 0
)
BETA_TIME_PT_CL = Beta(
    'BETA_TIME_PT_CL', choice_betas['BETA_TIME_PT_CL'], None, None, 0
)
BETA_WAITING_TIME = Beta(
    'BETA_WAITING_TIME', choice_betas['BETA_WAITING_TIME'], None, None, 0
)

# %%
# Definition of the utility functions.

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# %%
# Associate utility functions with the numbering of alternatives.
V = {0: V0, 1: V1, 2: V2}

# %%
# Conditional on omega, we have a logit model (called the kernel) for
# the choice.
condprob = models.logit(V, None, Choice)

# %%
# Conditional on omega, we have the product of ordered probit for the
# indicators.
condlike = (
    P_Envir01
    * P_Envir02
    * P_Envir03
    * P_Mobil11
    * P_Mobil14
    * P_Mobil16
    * P_Mobil17
    * condprob
)

# %%
# We integrate over omega using numerical integration.
loglike = log(Integrate(condlike * density, 'omega'))

# %%
# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'b05latent_choice_full'

# %%
# If estimation results are saved on file, we read them to speed up the process.
# If not, we estimate the parameters.
results = read_or_estimate(the_biogeme=the_biogeme, directory='saved_results')

# %%
print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')

# %%
results.getEstimatedParameters()



"""Deprecated expression, for backward compatibility

Michel Bierlaire
Thu Apr 3 09:55:57 2025
"""

import warnings

from .binary_max import BinaryMax
from .binary_min import BinaryMin
from .draws import Draws
from .integrate import IntegrateNormal
from .linear_utility import LinearUtility
from .multiple_sum import MultipleSum
from .normalcdf import NormalCdf


def deprecated_wrapper(old_name, new_class, comment=None):
    def wrapper(*args, **kwargs):
        message = (
            f"'{old_name}' is deprecated and will be removed in a future version. "
            f"Use '{new_class.__name__}' instead."
        )
        if comment:
            message += f"\n{comment}"
        warnings.warn(message, DeprecationWarning, stacklevel=2)
        return new_class(*args, **kwargs)

    wrapper.__name__ = old_name
    doc = (
        f"Deprecated wrapper for :class:`{new_class.__name__}`.\n\n"
        f".. warning::\n   This function is deprecated and will be removed in a future version.\n"
        f"   Use :class:`{new_class.__name__}` instead.\n"
    )
    if comment:
        doc += f"\n   {comment}\n"
    wrapper.__doc__ = doc
    return wrapper


bioLinearUtility = deprecated_wrapper("bioLinearUtility", LinearUtility)
bioMultSum = deprecated_wrapper("bioMultSum", MultipleSum)
bioDraws = deprecated_wrapper("bioDraws", Draws)
bioNormalCdf = deprecated_wrapper("bioNormalCdf", NormalCdf)
bioMin = deprecated_wrapper("bioMin", BinaryMin)
bioMax = deprecated_wrapper("bioMax", BinaryMax)
comment_integrate = (
    'In Biogeme 3.2, Integrate calculated the integral from -infinity to +infinity of f(x) dx.\n'
    'Since Biogeme 3.3, it is not available anymore.\n'
    'Instead, IntegrateNormal calculates the integral from -infinity to +infinity of f(x) * phi(x) dx,\n'
    'where phi(x) is the probability density function of the normal distribution.'
)
Integrate = deprecated_wrapper("Integrate", IntegrateNormal, comment=comment_integrate)



"""
Generates a F12 output for ALOGIT

Michel Bierlaire
Thu Oct 3 10:09:52 2024
"""

import datetime
import logging
import os

from biogeme.version import get_version

from .estimation_results import (
    EstimateVarianceCovariance,
    EstimationResults,
    calculates_correlation_matrix,
)
from .pandas_output import get_pandas_estimated_parameters

logger = logging.getLogger(__name__)


def get_f12(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance = EstimateVarianceCovariance.ROBUST,
) -> str:
    """F12 is a format used by the software ALOGIT to
    report estimation results.

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :return: results formatted in F12 format
    """

    covar_header = str(variance_covariance_type)

    # checkline1 = (
    #    '0000000001111111111222222222233333333334444444444'
    #    '5555555555666666666677777777778'
    # )
    # checkline2 = (
    #    '1234567890123456789012345678901234567890123456789'
    #    '0123456789012345678901234567890'
    # )

    results = ''

    # results += f'{checkline1}\n'
    # results += f'{checkline2}\n'

    # Line 1, title, characters 1-79
    results += f'{estimation_results.model_name[:79]: >79}\n'

    # Line 2, subtitle, characters 1-27, and time-date, characters 57-77
    t = f'From biogeme {get_version()}'
    d = f'{datetime.datetime.now()}'[:19]
    results += f'{t[:27]: <56}{d: <21}\n'

    # Line 3, "END" (this is historical!)
    results += 'END\n'

    # results += f'{checkline1}\n'
    # results += f'{checkline2}\n'

    # Line 4-(K+3), coefficient values
    #  characters 1-4, "   0" (again historical)
    #  characters 6-15, coefficient label, suggest using first 10
    #      characters of label in R
    #  characters 16-17, " F" (this indicates whether or not the
    #      coefficient is constrained)
    #  characters 19-38, coefficient value   20 chars
    #  characters 39-58, standard error      20 chars

    # mystats = estimation_results.get_general_statistics()
    table = get_pandas_estimated_parameters(
        estimation_results=estimation_results,
        variance_covariance_type=variance_covariance_type,
    )
    parameters_indices = table.index.to_list()
    for parameter_index in parameters_indices:
        values = table.loc[parameter_index]
        name = values['Name']
        results += '   0 '
        results += f'{name[:10]: >10}'
        if 'Active bound' in values:
            if values['Active bound'] == 1:
                results += ' T'
            else:
                results += ' F'
        else:
            results += ' F'
        results += ' '
        results += f' {values["Value"]: >+19.12e}'
        column_name = f'{covar_header} std err.'
        results += f' {values[column_name]: >+19.12e}'
        results += '\n'

    # Line K+4, "  -1" indicates end of coefficients
    results += '  -1\n'

    # results += f'{checkline1}\n'
    # results += f'{checkline2}\n'

    # Line K+5, statistics about run
    #   characters 1-8, number of observations        8 chars
    #   characters 9-27, likelihood-with-constants   19 chars
    #   characters 28-47, null likelihood            20 chars
    #   characters 48-67, final likelihood           20 chars

    results += f'{estimation_results.sample_size: >8}'
    # The cte log likelihood is not available. We put 0 instead.
    results += f' {0: >18}'
    if estimation_results.null_log_likelihood is not None:
        results += f' {estimation_results.null_log_likelihood: >+19.12e}'
    else:
        results += f' {0: >19}'
    results += f' {estimation_results.final_log_likelihood: >+19.12e}'
    results += '\n'

    # results += f'{checkline1}\n'
    # results += f'{checkline2}\n'

    # Line K+6, more statistics
    #   characters 1-4, number of iterations (suggest use 0)        4 chars
    #   characters 5-8, error code (please use 0)                   4 chars
    #   characters 9-29, time and date (sugg. repeat from line 2)  21 chars

    if "Number of iterations" in estimation_results.optimization_messages:
        results += (
            f'{estimation_results.optimization_messages["Number of iterations"]: >4}'
        )
    else:
        results += f'{0: >4}'
    results += f'{0: >4}'
    results += f'{d: >21}'
    results += '\n'

    # results += f'{checkline1}\n'
    # results += f'{checkline2}\n'

    # Lines (K+7)-however many we need, correlations*100000
    #   10 per line, fields of width 7
    #   The order of these is that correlation i,j (i>j) is in position
    #   (i-1)*(i-2)/2+j, i.e.
    #   (2,1) (3,1) (3,2) (4,1) etc.

    count = 0
    variance_covariance_matrix = estimation_results.get_variance_covariance_matrix(
        variance_covariance_type=variance_covariance_type
    )
    correlation_matrix = calculates_correlation_matrix(
        covariance=variance_covariance_matrix
    )
    for i, coefi in enumerate(estimation_results.beta_names):
        for j in range(0, i):
            try:
                corr = int(100000 * correlation_matrix[i][j])
            except OverflowError:
                corr = 999999
            results += f'{corr:7d}'
            count += 1
            if count % 10 == 0:
                results += '\n'
    results += '\n'
    return results


def generate_f12_file(
    estimation_results: EstimationResults,
    filename: str,
    overwrite=False,
    variance_covariance_type: EstimateVarianceCovariance | None = None,
) -> None:
    """Generate a F12 file with the estimation results

    :param estimation_results: estimation results
    :param filename: name of the file
    :param overwrite: if True and the file exists, it is overwritten
    :param variance_covariance_type: select which type of variance-covariance matrix is used to generate the
        statistics. If None, the bootstrap one is used if available. If not available, the robust one.
    """
    if variance_covariance_type is None:
        variance_covariance_type = (
            estimation_results.get_default_variance_covariance_matrix()
        )
    if (
        variance_covariance_type == EstimateVarianceCovariance.BOOTSTRAP
        and estimation_results.bootstrap_time is None
    ):
        logger.warning(
            f'No bootstrap data is available. The robust variance-covariance matrix is used instead.'
        )
        variance_covariance_type = EstimateVarianceCovariance.ROBUST

    if not overwrite and os.path.exists(filename):
        raise FileExistsError(f"The file '{filename}' already exists.")

    with open(filename, 'w') as file:
        content = get_f12(
            estimation_results=estimation_results,
            variance_covariance_type=variance_covariance_type,
        )
        print(content, file=file)
    logger.info(f'File {filename} has been generated.')



"""

Various integration methods
===========================

Calculation of a simple integral using numerical integration and
Monte-Carlo integration with various types of draws, including Halton
draws base 13. It illustrates how to use draws that are not directly
available in Biogeme.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:46:01 2023
"""

import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import draws
from biogeme.expressions import exp, bioDraws, MonteCarlo

# %%
# We create a fake database with one entry, as it is required
# to store the draws.
df = pd.DataFrame()
df['FakeColumn'] = [1.0]
database = db.Database('fakeDatabase', df)


# %%
def halton13(sample_size: int, number_of_draws: int) -> np.array:
    """
    The user can define new draws. For example, Halton draws
    with base 13, skipping the first 10 draws.

    :param sample_size: number of observations for which draws must be
                       generated.
    :param number_of_draws: number of draws to generate.

    """
    return draws.getHaltonDraws(sample_size, number_of_draws, base=13, skip=10)


# %%
mydraws = {'HALTON13': (halton13, 'Halton draws, base 13, skipping 10')}
database.setRandomNumberGenerators(mydraws)

# %%
# Integrate with pseudo-random number generator.
integrand = exp(bioDraws('U', 'UNIFORM'))
simulatedI = MonteCarlo(integrand)

# %%
# Integrate with Halton draws, base 2.
integrand_halton = exp(bioDraws('U_halton', 'UNIFORM_HALTON2'))
simulatedI_halton = MonteCarlo(integrand_halton)

# %%
# Integrate with Halton draws, base 13.
integrand_halton13 = exp(bioDraws('U_halton13', 'HALTON13'))
simulatedI_halton13 = MonteCarlo(integrand_halton13)

# %%
# Integrate with MLHS.
integrand_mlhs = exp(bioDraws('U_mlhs', 'UNIFORM_MLHS'))
simulatedI_mlhs = MonteCarlo(integrand_mlhs)

# %%
# True value
trueI = exp(1.0) - 1.0

# %%
# Number of draws.
R = 20000

# %%
sampleVariance = MonteCarlo(integrand * integrand) - simulatedI * simulatedI
stderr = (sampleVariance / R) ** 0.5
error = simulatedI - trueI

# %%
sampleVariance_halton = (
    MonteCarlo(integrand_halton * integrand_halton)
    - simulatedI_halton * simulatedI_halton
)
stderr_halton = (sampleVariance_halton / R) ** 0.5
error_halton = simulatedI_halton - trueI

# %%
sampleVariance_halton13 = (
    MonteCarlo(integrand_halton13 * integrand_halton13)
    - simulatedI_halton13 * simulatedI_halton13
)
stderr_halton13 = (sampleVariance_halton13 / R) ** 0.5
error_halton13 = simulatedI_halton13 - trueI

# %%
sampleVariance_mlhs = (
    MonteCarlo(integrand_mlhs * integrand_mlhs) - simulatedI_mlhs * simulatedI_mlhs
)
stderr_mlhs = (sampleVariance_mlhs / R) ** 0.5
error_mlhs = simulatedI_mlhs - trueI

# %%
simulate = {
    'Analytical Integral': trueI,
    'Simulated Integral': simulatedI,
    'Sample variance   ': sampleVariance,
    'Std Error         ': stderr,
    'Error             ': error,
    'Simulated Integral (Halton)': simulatedI_halton,
    'Sample variance (Halton)   ': sampleVariance_halton,
    'Std Error (Halton)         ': stderr_halton,
    'Error (Halton)             ': error_halton,
    'Simulated Integral (Halton13)': simulatedI_halton13,
    'Sample variance (Halton13)   ': sampleVariance_halton13,
    'Std Error (Halton13)         ': stderr_halton13,
    'Error (Halton13)             ': error_halton13,
    'Simulated Integral (MLHS)': simulatedI_mlhs,
    'Sample variance (MLHS)   ': sampleVariance_mlhs,
    'Std Error (MLHS)         ': stderr_mlhs,
    'Error (MLHS)             ': error_mlhs,
}

# %%
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b02simple_integral'
results = biosim.simulate(theBetaValues={})
results

# %%
# Reorganize the results.
print(f'Analytical integral: {results.iloc[0]["Analytical Integral"]:.6g}')
print(f"         \t{'Uniform':>15}\t{'Halton':>15}\t{'Halton13':>15}\t{'MLHS':>15}")
print(
    f'Simulated\t{results.iloc[0]["Simulated Integral"]:>15.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton)"]:>15.6g}\t'
    f'{results.iloc[0]["Simulated Integral (Halton13)"]:>15.6g}\t'
    f'{results.iloc[0]["Simulated Integral (MLHS)"]:>15.6g}'
)
print(
    f'Sample var.\t{results.iloc[0]["Sample variance   "]:>15.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton)   "]:>15.6g}\t'
    f'{results.iloc[0]["Sample variance (Halton13)   "]:>15.6g}\t'
    f'{results.iloc[0]["Sample variance (MLHS)   "]:>15.6g}'
)
print(
    f'Std error\t{results.iloc[0]["Std Error         "]:>15.6g}\t'
    f'{results.iloc[0]["Std Error (Halton)         "]:>15.6g}\t'
    f'{results.iloc[0]["Std Error (Halton13)         "]:>15.6g}\t'
    f'{results.iloc[0]["Std Error (MLHS)         "]:>15.6g}'
)
print(
    f'Error\t\t{results.iloc[0]["Error             "]:>15.6g}\t'
    f'{results.iloc[0]["Error (Halton)             "]:>15.6g}\t'
    f'{results.iloc[0]["Error (Halton13)             "]:>15.6g}\t'
    f'{results.iloc[0]["Error (MLHS)             "]:>15.6g}'
)



"""

Combine many specifications: exception is raised
================================================

We combine many specifications, defined in :ref:`everything_spec_section`.
This leads to a total of 432 specifications.
When the total number of specifications exceeds 100, Biogeme raises an
ValueOutOfRange exception when the estimate_catalog is called.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.

:author: Michel Bierlaire, EPFL
:date: Fri Jul 14 09:55:33 2023

"""
import biogeme.biogeme as bio
from biogeme.exceptions import BiogemeError

# %%
# See :ref:`everything_spec_section`.
from everything_spec import model_catalog, database

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b06everything'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Estimate the parameters.
# It does not work as there are two many specifications
try:
    dict_of_results = the_biogeme.estimate_catalog()
except BiogemeError as e:
    print(e)



"""

Simulation of panel model
=========================

Calculates each contribution to the log likelihood function using
simulation. We also calculate the individual parameters.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:16:29 2023

"""

import sys
import pickle
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.exceptions as excep
import biogeme.results as res
from biogeme.expressions import (
    Beta,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
)

# %%
# See the data processing script: :ref:`swissmetro_panel`.
from swissmetro_panel import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

print(f'Samples size = {database.getSampleSize()}')

# %%
# We use a low number of draws, as the objective is to illustrate the
# syntax. In practice, this value is insufficient to have a good
# approximation of the integral.
NUMBER_OF_DRAWS = 50

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b13panel_simul.py')

# %%
# Parameters.
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_ANTI')

# %%
# We do the same for the constants, to address serial correlation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_CAR_S = Beta('ASC_CAR_S', 1, None, None, 0)
ASC_CAR_RND = ASC_CAR + ASC_CAR_S * bioDraws('ASC_CAR_RND', 'NORMAL_ANTI')

ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_TRAIN_S = Beta('ASC_TRAIN_S', 1, None, None, 0)
ASC_TRAIN_RND = ASC_TRAIN + ASC_TRAIN_S * bioDraws('ASC_TRAIN_RND', 'NORMAL_ANTI')

ASC_SM = Beta('ASC_SM', 0, None, None, 1)
ASC_SM_S = Beta('ASC_SM_S', 1, None, None, 0)
ASC_SM_RND = ASC_SM + ASC_SM_S * bioDraws('ASC_SM_RND', 'NORMAL_ANTI')

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN_RND + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM_RND + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR_RND + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional on the random parameters, the likelihood of one observation is
# given by the logit model (called the kernel).
obsprob = models.logit(V, av, CHOICE)

# %%
# Conditional on the random parameters, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# %%
# We integrate over the random parameters using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# %%
# We retrieve the parameters estimates.
try:
    results = res.bioResults(pickleFile='saved_results/b12panel.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b12panel.py '
        'in order to generate the '
        'file b12panel.pickle.'
    )

# %%
# Simulate to recalculate the log likelihood directly from the
# formula, without the Biogeme object
simulated_loglike = logprob.getValue_c(
    database=database,
    betas=results.getBetaValues(),
    numberOfDraws=NUMBER_OF_DRAWS,
    aggregation=True,
    prepareIds=True,
)

# %%
print(f'Simulated log likelihood: {simulated_loglike}')

# %%
# We also calculate the individual parameters for the time coefficient.
numerator = MonteCarlo(B_TIME_RND * condprobIndiv)
denominator = MonteCarlo(condprobIndiv)

simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
}

# %%
# Creation of the Biogeme object.
biosim = bio.BIOGEME(database, simulate, parameter_file='few_draws.toml')

# %%
# Suimulation.
sim = biosim.simulate(results.getBetaValues())

# %%
sim['Individual-level parameters'] = sim['Numerator'] / sim['Denominator']

# %%
print(f'{sim.shape=}')

# %%
sim



"""
Factory tools for managing random draw generation in Biogeme.

This module defines classes for encapsulating and orchestrating
the creation of random draws used in simulation-based estimation.

It distinguishes between native and user-defined draw generators,
validates inputs, and constructs a final tensor of draws with shape:
    (sample_size, number_of_draws, number_of_variables)

Michel Bierlaire
Wed Mar 26 19:30:36 2025
"""

from dataclasses import dataclass
from collections.abc import Callable

import numpy as np

from biogeme.exceptions import BiogemeError
from .native_draws import (
    native_random_number_generators,
    RandomNumberGeneratorTuple,
)


@dataclass
class DrawSpec:
    """
    Encapsulates the configuration for generating draws for a specific variable.

    :param name: Name of the variable that requires simulated draws.
    :param draw_type: Identifier for the type of draw (native or user-defined).
    :param generator: A callable that takes (sample_size, number_of_draws)
                      and returns a NumPy array of draws.
    """

    name: str
    draw_type: str
    generator: Callable[[int, int], np.ndarray]


class DrawFactory:
    """
    Manages native and user-defined random draw generators and builds
    draw specifications and arrays for multiple variables.

    This class is useful for transforming a mapping of variable names and draw types
    into a single array of random numbers ready for use in simulation-based estimation.
    """

    def __init__(self, user_generators: dict[str, RandomNumberGeneratorTuple] | None):
        """
        Initializes the manager and validates user-defined generators.

        :param user_generators: A dictionary of user-defined draw types.
        :raises ValueError: if any user generator name collides with native keywords.
        """
        self.native_generators = native_random_number_generators
        self.user_generators = {} if user_generators is None else user_generators
        self._validate_reserved_keywords()

    def _validate_reserved_keywords(self) -> None:
        """Ensure no user-defined generator overrides a native one."""
        for key in self.native_generators:
            if key in self.user_generators:
                raise ValueError(
                    f"{key} is a reserved keyword and cannot be used "
                    "as a user-defined draw generator name."
                )

    def get_generator(
        self, draw_type: str, name: str
    ) -> Callable[[int, int], np.ndarray]:
        """
        Retrieves a draw generator function based on the requested type.

        :param draw_type: Type identifier of the draw (e.g., 'UNIFORM', 'HALTON').
        :param name: Variable name (used for error context).
        :return: A function that generates a NumPy array of draws.
        :raises BiogemeError: if the draw type is not recognized.
        """
        if draw_type in self.native_generators:
            return self.native_generators[draw_type].generator
        if draw_type in self.user_generators:
            return self.user_generators[draw_type].generator
        raise BiogemeError(
            f"Unknown draw type '{draw_type}' for variable '{name}'. "
            f"Available native types: {list(self.native_generators.keys())}. "
            f"User-defined types: {list(self.user_generators.keys())}."
        )

    def make_draw_specs(
        self, draw_types: dict[str, str], variable_names: list[str]
    ) -> list[DrawSpec]:
        """
        Generates a list of DrawSpec objects for each variable.

        :param draw_types: Mapping from variable name to draw type.
        :param variable_names: List of variable names requiring simulated draws.
        :return: List of fully constructed DrawSpec objects.
        """
        return [
            DrawSpec(
                name=name,
                draw_type=draw_types[name],
                generator=self.get_generator(draw_types[name], name),
            )
            for name in variable_names
        ]

    def generate_draws(
        self,
        draw_types: dict[str, str],
        variable_names: list[str],
        sample_size: int,
        number_of_draws: int,
    ) -> np.ndarray:
        """
        Generates a 3D NumPy array of draws for all specified variables.

        :param draw_types: Mapping from variable name to draw type.
        :param variable_names: Ordered list of variable names.
        :param sample_size: Number of observations in the sample.
        :param number_of_draws: Number of Monte Carlo draws per observation.
        :return: A NumPy array of shape (sample_size, number_of_draws, len(variable_names)).
        :raises BiogemeError: if any generator returns a mis-shaped array.
        """
        specs = self.make_draw_specs(draw_types, variable_names)
        draws = []
        for spec in specs:
            array = spec.generator(sample_size, number_of_draws)
            self._check_shape(spec.name, array, sample_size, number_of_draws)
            draws.append(array)
        return np.moveaxis(np.array(draws), 0, -1)

    def _check_shape(self, name, array, sample_size, number_of_draws):
        """
        Verifies that the shape of a generated draw array matches expectations.

        :param name: Variable name.
        :param array: Array returned by the draw generator.
        :param sample_size: Expected number of rows.
        :param number_of_draws: Expected number of columns.
        :raises BiogemeError: if the shape is invalid.
        """
        if array.shape != (sample_size, number_of_draws):
            raise BiogemeError(
                f"Draws for '{name}' must have shape ({sample_size}, {number_of_draws}), "
                f"but got {array.shape}."
            )



"""Module in charge of functionalities related to the choice set generation


    For thew main sample, all alternatives except the last one must be used: 0 to
    J-1.  For MEV models, the approximation of the sum capturing the
    nests requires another sample not based on the choice.

:author: Michel Bierlaire
:date: Fri Oct 27 12:50:06 2023
"""

import copy
import logging
import os

import pandas as pd
from tqdm import tqdm

from biogeme.database import Database
from biogeme.expressions import (
    Expression,
    list_of_variables_in_expression,
    rename_all_variables,
)
from .sampling_context import MEV_PREFIX, SamplingContext
from .sampling_of_alternatives import SamplingOfAlternatives

tqdm.pandas()

logger = logging.getLogger(__name__)


class ChoiceSetsGeneration:
    """Class in charge of generating the choice sets for each individual."""

    def __init__(self, context: SamplingContext):
        """Constructor

        :param context: contains all the information that is needed to
            perform the sampling of alternatives.

        """

        self.sampling_of_alternatives = SamplingOfAlternatives(context)

        self.alternatives = context.alternatives
        self.individuals = context.individuals
        self.choice_column = context.choice_column
        self.number_of_individuals = context.number_of_individuals
        self.id_column = context.id_column
        self.partition = context.sampling_protocol
        self.second_partition = context.mev_sampling_protocol
        self.combined_variables = context.combined_variables
        self.biogeme_file_name = context.biogeme_file_name
        self.total_sample_size = context.total_sample_size
        self.second_sample_size = context.total_mev_sample_size
        self.cnl_nests = context.cnl_nests
        self.biogeme_data = None

    def get_attributes_from_expression(self, expression: Expression) -> set[str]:
        """Extract the names of the attributes of alternatives from an expression"""
        variables = {
            var.name
            for var in list_of_variables_in_expression(the_expression=expression)
        }
        attributes = set(self.alternatives.columns)
        return variables & attributes

    def process_row(self, individual_row: pd.Series) -> dict:
        """Process one row of the individual database

        :param individual_row: row corresponding to one individual

        :return: a dictionary containing the data for the extended row
        """
        choice = individual_row[self.choice_column]

        first_sample = self.sampling_of_alternatives.sample_alternatives(chosen=choice)

        # Create the columns
        flattened_first_series: pd.Series[float, tuple[int, str]] = first_sample.stack()
        flattened_first_dict = {
            f'{col_name}_{row}': value
            for (row, col_name), value in flattened_first_series.items()
        }

        row_data = individual_row.to_dict()
        row_data.update(flattened_first_dict)

        if self.second_partition is not None:
            second_sample = self.sampling_of_alternatives.sample_mev_alternatives()

            # Rename columns for second_sample without multi-level index
            flattened_second_series = second_sample.stack()

            flattened_second_dict = {
                f"{MEV_PREFIX}{col_name}_{row}": value
                for (row, col_name), value in flattened_second_series.items()
            }
            row_data.update(flattened_second_dict)

        return row_data

    def define_new_variables(self, database: Database):
        """Create the new variables

        :param database: database, in Biogeme format.
        """
        total_iterations = len(self.combined_variables) * self.total_sample_size
        with tqdm(
            total=total_iterations, desc="Defining new variables..."
        ) as progress_bar:
            for new_variable in self.combined_variables:
                for index in range(self.total_sample_size):
                    copy_expression = copy.deepcopy(new_variable.formula)
                    attributes = self.get_attributes_from_expression(copy_expression)
                    rename_all_variables(
                        expr=copy_expression,
                        old_name=new_variable.name,
                        new_name=f'{new_variable.name}_{index}',
                    )
                    database.define_variable(
                        f"{new_variable.name}_{index}", copy_expression
                    )
                    progress_bar.update(1)
                if self.second_partition is not None:
                    for index in range(self.second_sample_size):
                        copy_expression = copy.deepcopy(new_variable.formula)
                        attributes = self.get_attributes_from_expression(
                            copy_expression
                        )
                        rename_all_variables(
                            expr=copy_expression,
                            old_name=new_variable.name,
                            new_name=f'{MEV_PREFIX}{new_variable.name}_{index}',
                        )

    def sample_and_merge(self, recycle: bool = False) -> Database:
        """Loops on the individuals and generate a choice set for each of them

        :param recycle: if True, if the data file already exists, it is not re-created.

        :return: database for Biogeme
        """
        if recycle:
            if os.path.exists(self.biogeme_file_name):
                biogeme_data = pd.read_csv(self.biogeme_file_name)
                biogeme_database = Database("merged_data", biogeme_data)
                return biogeme_database
            warning_msg = f"File {self.biogeme_file_name} does not exist."
            logger.warning(warning_msg)

        size = (
            f'{self.total_sample_size}'
            if self.second_sample_size is None
            else f'{self.total_sample_size} + {self.second_sample_size}'
        )
        logger.info(
            f"Generating {size} alternatives for "
            f"{self.number_of_individuals} observations"
        )
        biogeme_data = self.individuals.progress_apply(
            self.process_row, axis=1, result_type="expand"
        )
        biogeme_database = Database("merged_data", biogeme_data)
        logger.info("Define new variables")
        self.define_new_variables(biogeme_database)
        biogeme_data.to_csv(self.biogeme_file_name, index=False)
        logger.info(f"File {self.biogeme_file_name} has been created.")
        return biogeme_database



"""
Generates the estimation results in Pandas

Michel Bierlaire
Wed Oct 2 06:43:33 2024
"""

import logging

import numpy as np
import pandas as pd

from .estimation_results import (
    EstimateVarianceCovariance,
    EstimationResults,
    calc_p_value,
    calculates_correlation_matrix,
)
from ..exceptions import BiogemeError

logger = logging.getLogger(__name__)


def get_pandas_one_parameter(
    estimation_results: EstimationResults,
    parameter_index: int,
    variance_covariance_type: EstimateVarianceCovariance,
    parameter_number=None,
    parameter_name=None,
) -> dict[str, float | int | str]:
    """Generate one row of the Pandas table of the estimated parameters.

    :param estimation_results: estimation results.
    :param parameter_index: index of the parameter
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param parameter_number: number of the parameter to report. If None, it is the index.
    :param parameter_name: name of the parameter to report. If None, taken from estimation results.
    :return: one row of the table
    """
    if parameter_index < 0 or parameter_index >= len(estimation_results.beta_names):
        error_msg = f'Invalid parameter index {parameter_index}. Valid range: 0- {len(estimation_results.beta_names)-1}'
        raise ValueError(error_msg)
    if parameter_number is None:
        parameter_number = parameter_index
    if parameter_name is None:
        parameter_name = estimation_results.beta_names[parameter_index]

    covar_header = covar_header = str(variance_covariance_type)

    value = estimation_results.get_parameter_value_from_index(
        parameter_index=parameter_index
    )
    std_err = (
        estimation_results.get_parameter_std_err_from_index(
            parameter_index=parameter_index, estimate_var_covar=variance_covariance_type
        )
        if estimation_results.are_derivatives_available
        else np.nan
    )
    t_test = (
        estimation_results.get_parameter_t_test_from_index(
            parameter_index=parameter_index,
            estimate_var_covar=variance_covariance_type,
            target=0,
        )
        if estimation_results.are_derivatives_available
        else np.nan
    )
    p_value = (
        estimation_results.get_parameter_p_value_from_index(
            parameter_index=parameter_index,
            estimate_var_covar=variance_covariance_type,
            target=0,
        )
        if estimation_results.are_derivatives_available
        else np.nan
    )

    the_row = {
        '#': parameter_number,
        'Name': parameter_name,
        'Value': value,
        f'{covar_header} std err.': std_err,
        f'{covar_header} t-stat.': t_test,
        f'{covar_header} p-value': p_value,
    }
    if estimation_results.is_any_bound_active():
        the_row['Active bound'] = estimation_results.is_bound_active(
            parameter_name=estimation_results.beta_names[parameter_index]
        )
    return the_row


def get_pandas_estimated_parameters(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance | None = None,
    renumbering_parameters: dict[int, int] | None = None,
    renaming_parameters: dict[str, str] | None = None,
) -> pd.DataFrame:
    """Get the estimated parameters as a pandas data frame

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param renumbering_parameters: a dict that suggests new numbers for parameters
    :param renaming_parameters: a dict that suggests new names for some or all parameters.
    :param variance_covariance_type: select which type of variance-covariance matrix is used to generate the
        statistics. If None, the bootstrap one is used if available. If not available, the robust one.
    :return: a Pandas data frame
    """
    if variance_covariance_type is None:
        variance_covariance_type = (
            estimation_results.get_default_variance_covariance_matrix()
        )
    if (
        variance_covariance_type == EstimateVarianceCovariance.BOOTSTRAP
        and estimation_results.bootstrap_time is None
    ):
        logger.warning(
            f'No bootstrap data is available. The robust variance-covariance matrix is used instead.'
        )
        variance_covariance_type = EstimateVarianceCovariance.ROBUST
    if renumbering_parameters is not None:
        # Verify that the numbering is well defined
        number_values = list(renumbering_parameters.values())
        if len(number_values) != len(set(number_values)):
            error_msg = f'The new numbering cannot assign the same number to two different parameters.'
            raise BiogemeError(error_msg)

    if renaming_parameters is not None:
        # Verify that the renaming is well defined.
        name_values = list(renaming_parameters.values())
        if len(name_values) != len(set(name_values)):
            warning_msg = f'The new renaming assigns the same name for multiple parameters. It may not be the desired action.'
            logger.warning(warning_msg)
    all_rows = {}
    for parameter_index, parameter_name in enumerate(estimation_results.beta_names):
        new_number = (
            renumbering_parameters.get(parameter_index)
            if renumbering_parameters is not None
            else parameter_index
        )
        new_name = (
            renaming_parameters.get(parameter_name)
            if renaming_parameters is not None
            else estimation_results.beta_names[parameter_index]
        )

        the_row = get_pandas_one_parameter(
            estimation_results=estimation_results,
            parameter_index=parameter_index,
            variance_covariance_type=variance_covariance_type,
            parameter_number=new_number,
            parameter_name=new_name,
        )
        all_rows[new_number] = the_row

    list_of_all_rows = [all_rows[a_row_number] for a_row_number in sorted(all_rows)]
    the_frame = pd.DataFrame(list_of_all_rows)
    the_frame.set_index('#', inplace=True)
    the_frame.index.name = None

    return the_frame


def get_pandas_one_pair_of_parameters(
    estimation_results: EstimationResults,
    first_parameter_index: int,
    second_parameter_index: int,
    variance_covariance_type: EstimateVarianceCovariance,
    first_parameter_name=None,
    second_parameter_name=None,
) -> dict[str, float | int | str]:
    """Generate one row of the Pandas table of the correlation data for estimated parameters.

    :param estimation_results: estimation results.
    :param first_parameter_index: index of the first parameter
    :param second_parameter_index: index of the second parameter
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param first_parameter_name: name of the parameter to report. If None, taken from estimation results.
    :param second_parameter_name: name of the parameter to report. If None, taken from estimation results.
    :return: one row of the table
    """
    if first_parameter_index < 0 or first_parameter_index >= len(
        estimation_results.beta_names
    ):
        error_msg = (
            f'Invalid parameter index {first_parameter_index}. Valid range: 0-'
            f' {len(estimation_results.beta_names)-1}'
        )
        raise ValueError(error_msg)
    if second_parameter_index < 0 or second_parameter_index >= len(
        estimation_results.beta_names
    ):
        error_msg = (
            f'Invalid parameter index {second_parameter_index}. Valid range: 0-'
            f' {len(estimation_results.beta_names)-1}'
        )
        raise ValueError(error_msg)

    if first_parameter_name is None:
        first_parameter_name = estimation_results.beta_names[first_parameter_index]
    if second_parameter_name is None:
        second_parameter_name = estimation_results.beta_names[second_parameter_index]

    covar_header = str(variance_covariance_type)

    covariance_matrix = estimation_results.get_variance_covariance_matrix(
        variance_covariance_type=variance_covariance_type
    )
    correlation_matrix = calculates_correlation_matrix(covariance=covariance_matrix)
    covariance = covariance_matrix[first_parameter_index, second_parameter_index]
    correlation = correlation_matrix[first_parameter_index, second_parameter_index]
    t_test = estimation_results.calculate_test(
        first_parameter_index, second_parameter_index, covariance_matrix
    )
    p_value = calc_p_value(t_test)
    the_row = {
        'First parameter': first_parameter_name,
        'Second parameter': second_parameter_name,
        f'{covar_header} covariance': covariance,
        f'{covar_header} correlation': correlation,
        f'{covar_header} t-stat.': t_test,
        f'{covar_header} p-value': p_value,
    }
    return the_row


def get_pandas_correlation_results(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance,
    involved_parameters: dict[str, str] | None = None,
) -> pd.DataFrame:
    """Get the correlation results in a Pandas data frame

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param involved_parameters: a dict that identifies the parameters to involve, as assign them with a name for the
        reporting.
    :return: a Pandas data frame
    """
    if involved_parameters is None:
        list_of_parameters = {
            index: name for index, name in enumerate(estimation_results.beta_names)
        }
    else:
        list_of_parameters = {
            estimation_results.get_parameter_index(orig_name): new_name
            for orig_name, new_name in involved_parameters.items()
        }
    list_of_rows = []
    for first_parameter_index, first_parameter_name in list_of_parameters.items():
        for second_parameter_index, second_parameter_name in list_of_parameters.items():
            if first_parameter_index > second_parameter_index:
                the_row = get_pandas_one_pair_of_parameters(
                    estimation_results=estimation_results,
                    first_parameter_index=first_parameter_index,
                    second_parameter_index=second_parameter_index,
                    variance_covariance_type=variance_covariance_type,
                    first_parameter_name=first_parameter_name,
                    second_parameter_name=second_parameter_name,
                )
                list_of_rows.append(the_row)
    the_frame = pd.DataFrame(list_of_rows)
    return the_frame



"""Implements the Box-Cox model

:author: Michel Bierlaire
:date: Wed Oct 25 08:52:44 2023
"""

import logging
from biogeme.expressions import (
    Expression,
    Elem,
    Numeric,
    log,
    Beta,
    ExpressionOrNumeric,
)

logger = logging.getLogger(__name__)


def boxcox(x: ExpressionOrNumeric, ell: ExpressionOrNumeric) -> Expression:
    """Box-Cox transform

    .. math:: B(x, \\ell) = \\frac{x^{\\ell}-1}{\\ell}.

    It has the property that

    .. math:: \\lim_{\\ell \\to 0} B(x,\\ell)=\\log(x).

    To avoid numerical difficulties, if :math:`\\ell < 10^{-5}`,
    the McLaurin approximation is used:

    .. math:: \\log(x) + \\ell \\log(x)^2 + \\frac{1}{6} \\ell^2 \\log(x)^3
              + \\frac{1}{24} \\ell^3 \\log(x)^4.

    :param x: a variable to transform.
    :param ell: parameter of the transformation.

    :return: the Box-Cox transform
    """
    if isinstance(ell, Beta) and (ell.upper_bound is None or ell.lower_bound is None):
        warning_msg = (
            f'It is advised to set the bounds on parameter {ell.name}. '
            f'A value of -10 and 10 should be appropriate: Beta("{ell.name}", '
            f'{ell.init_value}, -10, 10, {ell.status})'
        )
        logger.warning(warning_msg)

    regular = (x**ell - 1.0) / ell
    mclaurin = (
        log(x)
        + ell * log(x) ** 2
        + ell**2 * log(x) ** 3 / 6.0
        + ell**3 * log(x) ** 4 / 24.0
    )
    close_to_zero = (ell < Numeric(1.0e-5)) * (ell > -Numeric(1.0e-5))
    smooth = Elem({0: regular, 1: mclaurin}, close_to_zero)
    return Elem({0: smooth, 1: Numeric(0)}, x == 0)



"""

Combine many specifications: assisted specification algorithm
=============================================================

We combine many specifications, defined in :ref:`everything_spec_section`. 
This leads to a total of 432 specifications.
The algorithm implemented in the AssistedSpecification object is used to
investigate some of these specifications.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.

:author: Michel Bierlaire, EPFL
:date: Sat Jul 15 15:02:20 2023

"""
from typing import Optional
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme.assisted import AssistedSpecification
from biogeme.multiobjectives import loglikelihood_dimension
from everything_spec import model_catalog, database
from biogeme.results import bioResults, compile_estimation_results

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07everything_assisted')

PARETO_FILE_NAME = 'b07everything_assisted.pareto'


# %%
# Function verifying that the estimation results are valid.
def validity(results: bioResults) -> tuple[bool, Optional[str]]:
    """Function verifying that the estimation results are valid.

    The results are not valid if any of the time or cost coefficient is non negative.
    """
    for beta in results.data.betas:
        if 'TIME' in beta.name and beta.value >= 0:
            return False, f'{beta.name} = {beta.value}'
        if 'COST' in beta.name and beta.value >= 0:
            return False, f'{beta.name} = {beta.value}'
    return True, None


# %%
# Create the Biogeme object
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b07everything'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Estimate the parameters using assisted specification algorithm.
assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
    validity=validity,
)

non_dominated_models = assisted_specification.run()

# %%
print(f'A total of {len(non_dominated_models)} models have been generated.')

# %%
compiled_results, specs = compile_estimation_results(
    non_dominated_models, use_short_names=True
)

# %%
compiled_results

# %%
# Glossary
for short_name, spec in specs.items():
    print(f'{short_name}\t{spec}')



"""

Illustration of additional features of Biogeme
==============================================

Same model as b01logit, using bioLinearUtility, segmentations
 and features.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:03:31 2023

"""
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.segmentation as seg
from biogeme.expressions import Beta, bioLinearUtility

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    GA,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
    SM_AV,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b01logit_bis.py')

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)

# %%
# Starting value.
# We use starting values estimated from a previous run
B_TIME = Beta('B_TIME', -1.28, None, None, 0)
B_COST = Beta('B_COST', -1.08, None, None, 0)

# %%
# Define segmentations.
gender_segmentation = database.generate_segmentation(
    variable=MALE, mapping={0: 'female', 1: 'male'}
)

GA_segmentation = database.generate_segmentation(
    variable=GA, mapping={0: 'without_ga', 1: 'with_ga'}
)

segmentations_for_asc = [
    gender_segmentation,
    GA_segmentation,
]

# %%
# Segmentation of the constants.
ASC_TRAIN_segmentation = seg.Segmentation(ASC_TRAIN, segmentations_for_asc)
segmented_ASC_TRAIN = ASC_TRAIN_segmentation.segmented_beta()
ASC_CAR_segmentation = seg.Segmentation(ASC_CAR, segmentations_for_asc)
segmented_ASC_CAR = ASC_CAR_segmentation.segmented_beta()

# %%
# Definition of the utility functions.
terms1 = [(B_TIME, TRAIN_TT_SCALED), (B_COST, TRAIN_COST_SCALED)]
V1 = segmented_ASC_TRAIN + bioLinearUtility(terms1)

terms2 = [(B_TIME, SM_TT_SCALED), (B_COST, SM_COST_SCALED)]
V2 = bioLinearUtility(terms2)

terms3 = [(B_TIME, CAR_TT_SCALED), (B_COST, CAR_CO_SCALED)]
V3 = segmented_ASC_CAR + bioLinearUtility(terms3)

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model.
#
# This is the contribution of each observation to the log likelihood
# function.
logprob = models.loglogit(V, av, CHOICE)

# %%
# User notes.
#
# These notes will be included as such in the report file.
USER_NOTES = (
    'Example of a logit model with three alternatives: Train, Car and'
    ' Swissmetro. Same as 01logit and '
    'introducing some options and features. In particular, bioLinearUtility,'
    ' and automatic segmentation of parameters.'
)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(
    database, logprob, userNotes=USER_NOTES, parameter_file='b01logit_bis.toml'
)

# %%
# Calculate the null log likelihood for reporting.
#
# As we have used starting values different from 0, the initial model
# is not the equal probability model.
the_biogeme.calculateNullLoglikelihood(av)
the_biogeme.modelName = 'b01logit_bis'

# %%
# Turn off saving iterations.
#
the_biogeme.saveIterations = False

# %%
# Estimate the parameters.
#
the_biogeme.bootstrap_samples = 100
results = the_biogeme.estimate(run_bootstrap=True)

# %%
# Get the results in a pandas table.
#
print('Parameters')
print('----------')
pandas_results = results.getEstimatedParameters()
pandas_results

# %%
# Get general statistics.
#
print('General statistics')
print('------------------')
stats = results.getGeneralStatistics()
for description, (value, formatting) in stats.items():
    print(f'{description}: {value:{formatting}}')

# %%
# Messages from the optimization algorithm.
#
print('Optimization algorithm')
print('----------------------')
for description, message in results.data.optimizationMessages.items():
    print(f'{description}:\t{message}')

# %%
# Generate the file in Alogit format.
#
results.writeF12(robustStdErr=True)
print(f'Estimation results in ALogit format generated: {results.data.F12FileName}')

# %%
# Generate LaTeX code with the results.
#
results.writeLaTeX()
print(f'Estimation results in LaTeX format generated: {results.data.latexFileName}')



"""

Estimation of several models
============================

Example of the estimation of several specifications of the model.

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:19:46 2023

"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, log
from biogeme.results import compile_estimation_results, pareto_optimal
from biogeme.catalog import Catalog, segmentation_catalogs

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
)

# %%
# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_CAR_MALE = Beta('ASC_CAR_MALE', 0, None, None, 0)
ASC_CAR_FEMALE = Beta('ASC_CAR_FEMALE', 0, None, None, 0)
ASC_TRAIN_MALE = Beta('ASC_TRAIN_MALE', 0, None, None, 0)
ASC_TRAIN_FEMALE = Beta('ASC_TRAIN_FEMALE', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
segmentation_gender = database.generate_segmentation(
    variable=MALE, mapping={0: 'female', 1: 'male'}
)

# %%
# We define catalogs with two different specifications for the
# ASC_CAR: non segmented, and segmented.
ASC_TRAIN_catalog, ASC_CAR_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_TRAIN, ASC_CAR],
    potential_segmentations=(segmentation_gender,),
    maximum_number=1,
)

# %%
# We now define a catalog  with the log travel time as well as the travel time.

# %%
# First for train
train_tt_catalog = Catalog.from_dict(
    catalog_name='train_tt_catalog',
    dict_of_expressions={
        'linear': TRAIN_TT_SCALED,
        'log': log(TRAIN_TT_SCALED),
    },
)

# %%
# Then for SM. But we require that the specification is the same as
# train by defining the same controller.
sm_tt_catalog = Catalog.from_dict(
    catalog_name='sm_tt_catalog',
    dict_of_expressions={
        'linear': SM_TT_SCALED,
        'log': log(SM_TT_SCALED),
    },
    controlled_by=train_tt_catalog.controlled_by,
)

# %%
# Definition of the utility functions with linear cost.
V1 = ASC_TRAIN_catalog + B_TIME * train_tt_catalog + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * sm_tt_catalog + B_COST * SM_COST_SCALED
V3 = ASC_CAR_catalog + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b20multiple_models'

# %%
dict_of_results = the_biogeme.estimate_catalog()

# %%
print(f'A total of {len(dict_of_results)} models have been estimated:')
for config, res in dict_of_results.items():
    print(f'{config}: LL={res.data.logLike:.2f} K={res.data.nparam}')

# %%
summary, description = compile_estimation_results(dict_of_results, use_short_names=True)
print(summary)

# %%
# Explanation of the names of the models.
for k, v in description.items():
    if k != v:
        print(f'{k}: {v}')

# %%
non_dominated_models = pareto_optimal(dict_of_results)
print(f'Out of them, {len(non_dominated_models)} are non dominated.')
for config, res in non_dominated_models.items():
    print(f'{config}')

# %%
summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=False
)
print(summary)



from .audit import (
    ChosenAvailable,
    audit_dataframe,
    audit_panel_dataframe,
    check_availability_of_chosen_alt,
    choice_availability_statistics,
)
from .container import Database
from .panel import PanelDatabase, RELEVANT_PREFIX, observation_suffix
from .sampling import sample_panel_with_replacement, sample_with_replacement



"""Arithmetic expressions accepted by Biogeme: sine

Michel Bierlaire
Fri Jun 27 2025, 17:27:22
"""

from __future__ import annotations

import logging

import jax.numpy as jnp
import numpy as np

from .base_expressions import ExpressionOrNumeric
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator

logger = logging.getLogger(__name__)


class sin(UnaryOperator):
    """
    sine expression
    """

    def __init__(self, child: ExpressionOrNumeric):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        super().__init__(child)

    def deep_flat_copy(self) -> sin:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(child=copy_child)

    def __str__(self) -> str:
        return f'sin({self.child})'

    def __repr__(self) -> str:
        return f'sin({repr(self.child)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        return np.sin(self.child.get_value())

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            child_value = child_jax(
                parameters, one_row, the_draws, the_random_variables
            )
            return jnp.sin(child_value)

        return the_jax_function



"""Implementation of the "translated" MDCEV model. See the technical report.

Michel Bierlaire
Tue Apr 9 09:12:55 2024
"""

import logging

import numpy as np

from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, log, Numeric, exp
from biogeme.mdcev.mdcev import Mdcev

logger = logging.getLogger(__name__)

MAX_EXP_ARGUMENT = np.log(np.finfo(dtype=float).max)


class Translated(Mdcev):
    """Implementation of the "translated" MDCEV model"""

    def __init__(
        self,
        model_name: str,
        baseline_utilities: dict[int, Expression],
        gamma_parameters: dict[int, Expression | None],
        alpha_parameters: dict[int, Expression] | None = None,
        scale_parameter: Expression | None = None,
        weights: Expression | None = None,
    ) -> None:
        """Implementation of the MDCEV model with Gamma profile."""
        super().__init__(
            model_name=model_name,
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameters,
            scale_parameter=scale_parameter,
            weights=weights,
        )

    def transformed_utility(
        self,
        the_id: int,
        the_consumption: Expression,
    ) -> Expression:
        """Calculates the utility for one alternative."""

        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)

        baseline_utility = self.baseline_utilities[the_id]

        alpha_parameter = self.alpha_parameters[the_id]

        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                baseline_utility
                + log(alpha_parameter)
                + (alpha_parameter - 1) * log(the_consumption)
            )

        return (
            baseline_utility
            + log(alpha_parameter)
            + (alpha_parameter - 1) * log(the_consumption + gamma)
        )

    def calculate_log_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the log of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return log(Numeric(1) - self.alpha_parameters[the_id]) - log(consumption)
        return log(Numeric(1) - self.alpha_parameters[the_id]) - log(
            consumption + gamma
        )

    def calculate_inverse_of_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the inverse of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return consumption / (Numeric(1) - self.alpha_parameters[the_id])
        return (consumption + gamma) / (Numeric(1) - self.alpha_parameters[the_id])

    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        baseline_utility: Expression = self.baseline_utilities[the_id]
        epsilon: Expression = (
            unscaled_epsilon
            if self.scale_parameter is None
            else unscaled_epsilon / self.scale_parameter
        )
        alpha: Expression = self.alpha_parameters[the_id]
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            log_result = baseline_utility + epsilon + alpha * log(the_consumption)
            return exp(log_result)
        log_result = baseline_utility + epsilon + alpha * log(the_consumption + gamma)
        return exp(log_result)

    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility used in the optimization problem solved for forecasting."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            if the_consumption == 0.0:
                return 0.0
            log_result = baseline_utility + epsilon + alpha * np.log(the_consumption)
            return np.exp(log_result)
        log_result = (
            baseline_utility
            + epsilon
            + alpha * np.log(the_consumption + gamma.get_value())
        )
        return np.exp(log_result)

    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting tp calculate the dual variable."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            log_result = (
                (
                    baseline_utility
                    + epsilon
                    + np.log(alpha)
                    + (alpha - 1) * np.log(the_consumption)
                )
                if the_consumption != 0.0
                else 0.0
            )
            return np.exp(log_result)
        log_result = (
            baseline_utility
            + epsilon
            + np.log(alpha)
            + (alpha - 1) * np.log(the_consumption + gamma.get_value())
        )
        return np.exp(log_result)

    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha = self.alpha_parameters[the_id].get_value()
        if np.isclose(alpha, 0) or np.isclose(alpha, 1.0):
            error_msg = f'Parameter alpha[{the_id}] must be strictly below 0 and 1: alpha[{the_id}={alpha:.3g}'
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]
        if dual_variable == 0.0:
            return 0.0
        log_ratio = np.log(dual_variable) - baseline_utility - epsilon - np.log(alpha)
        log_result = min(log_ratio / (alpha - 1), MAX_EXP_ARGUMENT)
        if gamma is None:
            return np.exp(log_result)
        return np.exp(log_result) - gamma.get_value()

    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        """Method providing model specific bounds on the dual variable. It not overloaded,
        default values are used.

        :param chosen_alternatives: list of alternatives that are chosen at the optimal solution
        :param one_observation: data for one observation.
        :param epsilon: draws from the error term.
        :return: a lower bound on the dual variable, such that the expenditure calculated for any larger value is
        well-defined and non negative.
        """
        return 0.0

    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        return []



"""Arithmetic expressions accepted by Biogeme: random variable for numerical integration

Michel Bierlaire
Fri Jun 27 2025, 14:50:37
"""

from __future__ import annotations

import logging

import jax.numpy as jnp
from biogeme.exceptions import BiogemeError

from .elementary_expressions import Elementary
from .elementary_types import TypeOfElementaryExpression
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class RandomVariable(Elementary):
    """
    Random variable for numerical integration
    """

    expression_type = TypeOfElementaryExpression.RANDOM_VARIABLE

    def __init__(self, name: str):
        """Constructor

        :param name: name of the random variable involved in the integration.
        :type name: string.
        """
        super().__init__(name)
        # Index of the random variable

    def deep_flat_copy(self) -> RandomVariable:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        return type(self)(name=self.name)

    def dict_of_elementary_expression(
        self, the_type: TypeOfElementaryExpression
    ) -> dict[str:Elementary]:
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression
        """
        if the_type == TypeOfElementaryExpression.RANDOM_VARIABLE:
            return {self.name: self}
        return {}

    @property
    def safe_rv_id(self) -> int:
        """Check the presence of the ID before using it"""
        if self.specific_id is None:
            raise BiogemeError(f"No id defined for random variable {self.name}")
        return self.specific_id

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.array:
            return jnp.take(the_random_variables, self.safe_rv_id, axis=-1)

        return the_jax_function



"""
Test the dict_of_formulas functions

:author: Michel Bierlaire
:date: Mon May 13 16:27:54 2024

"""

import unittest
from unittest.mock import patch, MagicMock

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression
from biogeme.dict_of_formulas import (
    check_validity,
    get_expression,
)


class TestExpressions(unittest.TestCase):

    def setUp(self):
        self.expression_mock = MagicMock(spec=Expression)
        self.dict_of_valid_loglike = {'log_like': self.expression_mock}
        self.dict_of_valid_weight = {'weight': self.expression_mock}

    def test_get_expression_valid(self):
        result = get_expression(
            dict_of_formulas=self.dict_of_valid_loglike, valid_keywords=['log_like']
        )
        self.assertEqual(result, self.expression_mock)

    def test_check_validity_valid(self):
        try:
            check_validity(self.dict_of_valid_loglike)
        except BiogemeError:
            self.fail("check_validity() raised BiogemeError unexpectedly!")

    def test_check_validity_invalid_type(self):
        with self.assertRaises(BiogemeError):
            check_validity({'log_like': 'not_an_expression'})

    @patch('biogeme.dict_of_formulas.logger')
    def test_get_expression_warning_similar_name(self, mock_logger):
        with patch('biogeme.dict_of_formulas.is_similar_to', return_value=True):
            get_expression(
                dict_of_formulas={'loc_like': self.expression_mock},
                valid_keywords=['log_like', 'loglike'],
            )
            self.assertEqual(mock_logger.warning.call_count, 2)

    @patch('biogeme.dict_of_formulas.logger')
    def test_get_expression_no_valid(self, mock_logger):
        result = get_expression(dict_of_formulas={}, valid_keywords=['weight'])
        self.assertIsNone(result)
        mock_logger.warning.assert_not_called()

    def test_get_expression_multiple_valid_names(self):
        dict_of_formulas = {
            'log_like': self.expression_mock,
            'loglike': self.expression_mock,
        }
        with self.assertRaises(BiogemeError):
            get_expression(
                dict_of_formulas=dict_of_formulas,
                valid_keywords=['log_like', 'loglike'],
            )


if __name__ == '__main__':
    unittest.main()



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    GA,
    INCOME,
    PURPOSE,
    SM_AV,
    SM_CO,
    TRAIN_AV_SP,
    TRAIN_CO,
    read_data,
)
from biogeme.database import Database, PanelDatabase
from biogeme.default_parameters import MISSING_VALUE
from biogeme.expressions import Beta, Draws, MonteCarlo, MultipleSum, Variable, exp, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

panel_database = PanelDatabase(database=database, panel_column='ID')

flat_dataframe, largest_group = panel_database.flatten_database(
    missing_data=MISSING_VALUE
)
flat_database = Database(name='flat_swissmetro', dataframe=flat_dataframe)

SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
SIGMA_CAR = Beta('SIGMA_CAR', 3.818728, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0.946202, None, None, 0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', 2.874614, None, None, 0)

EC_CAR = SIGMA_CAR * Draws('EC_CAR', 'NORMAL')
EC_SM = SIGMA_SM * Draws('EC_SM', 'NORMAL')
EC_TRAIN = SIGMA_TRAIN * Draws('EC_TRAIN', 'NORMAL')

ASC_CAR = Beta('ASC_CAR', -0.035345, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.747857, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -6.136095, None, 0, 0)
B_COST = Beta('B_COST', -2.955199, None, 0, 0)


# For latent class 1, where the time coefficient is zero
V11 = [
    ASC_TRAIN + B_COST * Variable(f'TRAIN_COST_SCALED__panel__{t:02d}') + EC_TRAIN
    for t in range(1, largest_group + 1)
]
V12 = [
    ASC_SM + B_COST * Variable(f'SM_COST_SCALED__panel__{t:02d}') + EC_SM
    for t in range(1, largest_group + 1)
]
V13 = [
    ASC_CAR + B_COST * Variable(f'CAR_CO_SCALED__panel__{t:02d}') + EC_CAR
    for t in range(1, largest_group + 1)
]

V1 = [{1: V11[t], 2: V12[t], 3: V13[t]} for t in range(largest_group)]

# For latent class 2, where the time coefficient is estimated
V21 = [
    ASC_TRAIN
    + B_TIME * Variable(f'TRAIN_TT_SCALED__panel__{t:02d}')
    + B_COST * Variable(f'TRAIN_COST_SCALED__panel__{t:02d}')
    + EC_TRAIN
    for t in range(1, largest_group + 1)
]
V22 = [
    ASC_SM
    + B_TIME * Variable(f'SM_TT_SCALED__panel__{t:02d}')
    + B_COST * Variable(f'SM_COST_SCALED__panel__{t:02d}')
    + EC_SM
    for t in range(1, largest_group + 1)
]
V23 = [
    ASC_CAR
    + B_TIME * Variable(f'CAR_TT_SCALED__panel__{t:02d}')
    + B_COST * Variable(f'CAR_CO_SCALED__panel__{t:02d}')
    + EC_CAR
    for t in range(1, largest_group + 1)
]

V2 = [{1: V21[t], 2: V22[t], 3: V23[t]} for t in range(largest_group)]

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# Class membership model
# Class membership model
CLASS_CTE = Beta('CLASS_CTE', -0.792233, None, None, 0)
CLASS_INC = Beta('CLASS_INC', -0.251995, None, None, 0)
W1 = CLASS_CTE + CLASS_INC * INCOME
prob_class1 = models.logit({1: W1, 2: 0}, None, 1)
prob_class2 = models.logit({1: W1, 2: 0}, None, 2)

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 1
obs_prob_1 = [
    models.loglogit(V1[t], av, Variable(f'CHOICE__panel__{t+1:02d}'))
    for t in range(largest_group)
]
prob1 = exp(MultipleSum(obs_prob_1))

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 2
obs_prob_2 = [
    models.loglogit(V2[t], av, Variable(f'CHOICE__panel__{t+1:02d}'))
    for t in range(largest_group)
]
prob2 = exp(MultipleSum(obs_prob_2))

# Conditional to the random variables, likelihood for the individual.
prob_indiv = prob_class1 * prob1 + prob_class2 * prob2

# We integrate over the random variables using Monte-Carlo
logprob = log(MonteCarlo(prob_indiv))


class Test16(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            flat_database,
            logprob,
            number_of_draws=100,
            seed=1111,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -3637.7836216145975, 2)


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: PanelLikelihoodTrajectory

Michel Bierlaire
11.04.2025 09:29
"""

from __future__ import annotations

import copy
import logging

from biogeme.exceptions import BiogemeError

from . import add_prefix_suffix_to_all_variables
from .base_expressions import Expression, ExpressionOrNumeric
from .conditional_sum import ConditionalSum, ConditionalTermTuple
from .exp import exp
from .jax_utils import JaxFunctionType
from .log import log
from .variable import Variable

logger = logging.getLogger(__name__)


class PanelLikelihoodTrajectory(Expression):
    """
    Likelihood of a sequences of observations for the same individual
    """

    def __init__(self, child: ExpressionOrNumeric):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        # In this case, the formula itself is not a child. It will have to be manipulated by the function
        # `set_maximum_number_of_observations_per_individual`. Therefore, the super().__init__() has no argument.
        super().__init__()
        self.child: Expression | None = None
        self.initial_formula: Expression = child
        self.maximum_number_of_observations_per_individual: int | None = None

    def deep_flat_copy(self) -> PanelLikelihoodTrajectory:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.initial_formula.deep_flat_copy()
        the_copy = PanelLikelihoodTrajectory(child=copy_child)
        if self.maximum_number_of_observations_per_individual is not None:
            the_copy.set_maximum_number_of_observations_per_individual(
                self.maximum_number_of_observations_per_individual
            )
        return the_copy

    def set_maximum_number_of_observations_per_individual(
        self, max_number: int
    ) -> None:
        from biogeme.database import observation_suffix, RELEVANT_PREFIX

        self.maximum_number_of_observations_per_individual = max_number
        if self.initial_formula.get_class_name() == 'exp':
            copies_of_expression = [
                copy.deepcopy(self.initial_formula.child)
                for _ in range(self.maximum_number_of_observations_per_individual)
            ]
        else:
            copies_of_expression = [
                log(copy.deepcopy(self.initial_formula))
                for _ in range(self.maximum_number_of_observations_per_individual)
            ]
        list_of_terms = []
        for index, a_copy in enumerate(copies_of_expression):
            suffix = observation_suffix(index)
            add_prefix_suffix_to_all_variables(expr=a_copy, prefix='', suffix=suffix)
            the_term = ConditionalTermTuple(
                condition=Variable(f'{RELEVANT_PREFIX}{suffix}'), term=a_copy
            )
            list_of_terms.append(the_term)

        self.child = exp(ConditionalSum(list_of_terms=list_of_terms))
        self.children.append(self.child)

    def __str__(self) -> str:
        return f'PanelLikelihoodTrajectory({self.initial_formula})'

    def __repr__(self) -> str:
        return f'PanelLikelihoodTrajectory({repr(self.initial_formula)})'

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates recursively a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        if self.child is None:
            error_msg = 'The PanelLikelihoodTrajectory has not been prepared before being evaluated.'
            raise BiogemeError(error_msg)
        if self.maximum_number_of_observations_per_individual is None:
            error_msg = (
                'Maximum number of observations per individual has not been defined.'
            )
            raise BiogemeError(error_msg)

        return self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )



"""Handles the segmentation of a single parameter using a discrete variable.

This class manages how a single Beta parameter is segmented according to a
discrete variable. It generates distinct parameters for each category in the
segmentation (except the reference), enabling models to estimate category-specific
effects.

Attributes:
    beta: The base Beta parameter being segmented.
    segmentation_tuple: The DiscreteSegmentationTuple defining the segmentation.
    variable: The discrete variable used for segmentation.
    reference: The reference category for the segmentation.
    mapping: Dictionary of category values to category names, excluding the reference.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .segmentation_context import DiscreteSegmentationTuple
from biogeme.exceptions import BiogemeError

if TYPE_CHECKING:
    from biogeme.expressions import Beta, Variable, Expression, Numeric


class OneSegmentation:

    def __init__(
        self,
        beta: Beta,
        segmentation_tuple: DiscreteSegmentationTuple,
    ):
        """
        Initialize the segmentation of a Beta parameter.

        Parameters:
            beta: The base Beta parameter to be segmented.
            segmentation_tuple: Describes the discrete segmentation including
                                the mapping of values to category names and the reference category.
        """
        self.beta: Beta = beta
        self.segmentation_tuple = segmentation_tuple
        self.variable: Variable = segmentation_tuple.variable
        self.reference: str = segmentation_tuple.reference
        self.mapping: dict[int, str] = {
            k: v for k, v in segmentation_tuple.mapping.items() if v != self.reference
        }

    def beta_name(self, category: str) -> str:
        """Construct the name of the parameter associated with a specific category

        :param category: name of the category
        :return: name of parameter for the category

        :raise BiogemeError: if the category is not listed in the
            mapping of the segmentation.
        """
        if category != self.reference and category not in self.mapping.values():
            error_msg = (
                f'Unknown category: {category}. List of known categories: '
                f'{self.mapping.values()}'
            )
            raise BiogemeError(error_msg)
        return (
            f'{self.beta.name}_ref'
            if category == self.reference
            else f'{self.beta.name}_diff_{category}'
        )

    def beta_expression(self, category: str | None = None) -> Beta:
        """Constructs the expression for the parameter associated with
            a specific category

        :param category: name of the category. If None, it is the reference category.
        :return: expression of the parameter for the category


        """
        from biogeme.expressions import Beta

        if category is None:
            category = self.reference
        name = self.beta_name(category)
        if category == self.reference:
            lower_bound = self.beta.lower_bound
            upper_bound = self.beta.upper_bound
        else:
            lower_bound = None
            upper_bound = None

        return Beta(
            name,
            self.beta.init_value,
            lower_bound,
            upper_bound,
            self.beta.status,
        )

    def beta_code(self, category: str, assignment: bool) -> str:
        """Constructs the Python code for the expression of the
            parameter associated with a specific category

        :param category: name of the category
        :param assignment: if True, the code includes the assignment to a variable.
        :return: the Python code
        """
        if category == self.reference:
            lower_bound = self.beta.lower_bound
            upper_bound = self.beta.upper_bound
        else:
            lower_bound = None
            upper_bound = None
        name = self.beta_name(category)
        if assignment:
            return (
                f"{name} = Beta('{name}', {self.beta.init_value}, "
                f"{lower_bound}, {upper_bound}, {self.beta.status})"
            )
        return (
            f"Beta('{name}', {self.beta.init_value}, {lower_bound}, "
            f"{upper_bound}, {self.beta.status})"
        )

    def list_of_expressions(self) -> list[Expression]:
        """Create a list of expressions involved in the segmentation of the parameter

        :return: list of expressions
        :rtype: list(biogeme.expressions.Expression)

        """
        from biogeme.expressions import Numeric

        terms = [
            self.beta_expression(category) * (self.variable == Numeric(value))
            for value, category in self.mapping.items()
        ]
        return terms

    def list_of_code(self) -> list[str]:
        """Create a list of Python codes for the expressions involved
            in the segmentation of the parameter

        :return: list of codes
        """
        return [
            (
                f"{self.beta_name(category)} "
                f"* (Variable('{self.variable.name}') == {value})"
            )
            for value, category in self.mapping.items()
        ]



"""Generation of models estimated with samples of alternatives

:author: Michel Bierlaire
:date: Fri Sep 22 12:14:59 2023
"""

import copy
import logging

from biogeme.expressions import (
    BelongsTo,
    ConditionalSum,
    ConditionalTermTuple,
    Expression,
    Variable,
    exp,
    log,
    logzero,
)
from biogeme.models import loglogit
from biogeme.nests import NestsForNestedLogit
from .sampling_context import CNL_PREFIX, LOG_PROBA_COL, MEV_WEIGHT, SamplingContext
from ..expressions.add_prefix_suffix import add_prefix_suffix_to_all_variables

logger = logging.getLogger(__name__)


class GenerateModel:
    """Class in charge of generating the biogeme expression for the
    loglikelihood function
    """

    def __init__(self, context: SamplingContext):
        """Constructor

        :param context: contains all the information that is needed to
            perform the sampling of alternatives.

        """

        self.context = context
        self.utility_function = context.utility_function
        self.total_sample_size = context.total_sample_size
        self.total_mev_sample_size = context.total_mev_sample_size
        self.attributes = context.attributes
        self.mev_prefix = context.mev_prefix

        self.utilities = {
            alt_id: self.generate_utility(prefix="", suffix=f"_{alt_id}")
            for alt_id in range(self.total_sample_size)
        }
        if self.context.mev_partition is None:
            self.mev_utilities = {
                alt_id: self.utilities[alt_id]
                for alt_id in range(1, self.total_sample_size)
            }
            logger.debug(
                f'No specific MEV partition. {self.total_sample_size} are sampled for MEV terms.'
            )
        else:
            self.mev_utilities = {
                alt_id: self.generate_utility(
                    prefix=self.mev_prefix, suffix=f"_{alt_id}"
                )
                for alt_id in range(self.context.total_mev_sample_size)
            }
            logger.debug(
                f'Specific MEV partition provided. {self.total_mev_sample_size} are sampled for MEV terms.'
            )

    def generate_utility(self, prefix: str, suffix: str) -> Expression:
        """Generate the utility function for one alternative

        :param prefix: prefix to add to the attributes

        :param suffix: suffix to add to the attributes

        """
        copy_utility = copy.deepcopy(self.utility_function)
        add_prefix_suffix_to_all_variables(
            expr=copy_utility, prefix=prefix, suffix=suffix
        )
        return copy_utility

    def get_logit(self) -> Expression:
        """Returns the expression for the log likelihood of the logit model"""

        corrected_utilities = {
            i: utility - Variable(f"{LOG_PROBA_COL}_{i}")
            for i, utility in self.utilities.items()
        }

        return loglogit(corrected_utilities, None, 0)

    def get_nested_logit(self, nests: NestsForNestedLogit) -> Expression:
        """Returns the expression for the log likelihood of the nested logit model

        :param nests: A tuple containing as many items as nests.
            Each item is also a tuple containing two items:

        - an object of type biogeme.expressions.expr.Expression representing
          the nest parameter,
        - a list containing the list of identifiers of the alternatives
          belonging to the nest.

        Example::

            nesta = MUA ,[1, 2, 3]
            nestb = MUB ,[4, 5, 6]
            nests = nesta, nestb


        """

        dict_of_mev_sums = {}

        # We first compute the MEV partial sum for each nest
        for nest in nests:
            mu_param = nest.nest_param
            list_of_alternatives = nest.list_of_alternatives
            self.context.check_valid_alternatives(set(list_of_alternatives))
            # We first build the MEV term using a sample of
            # alternatives. To build this term, we iterate from 1, as
            # we ignore the chosen alternative.
            list_of_terms = []
            for i, utility in self.mev_utilities.items():
                alternative_id = Variable(
                    f"{self.mev_prefix}{self.context.id_column}_{i}"
                )
                belong_to_nest = BelongsTo(alternative_id, set(list_of_alternatives))
                weight = Variable(f"{self.mev_prefix}{MEV_WEIGHT}_{i}")
                the_term = ConditionalTermTuple(
                    condition=belong_to_nest, term=weight * exp(mu_param * utility)
                )
                list_of_terms.append(the_term)
            dict_of_mev_sums[tuple(list_of_alternatives)] = ConditionalSum(
                list_of_terms
            )

        # We now add all relevant MEV terms to the utilities

        dict_of_mev_terms = {}
        for i, the_utility in self.utilities.items():
            alternative_id = Variable(f"{self.context.id_column}_{i}")
            list_of_terms = []
            for nest in nests:
                mu_param = nest.nest_param
                mev_sum = dict_of_mev_sums[tuple(nest.list_of_alternatives)]
                mev_term = (mu_param - 1.0) * the_utility + (
                    (1.0 / mu_param) - 1.0
                ) * log(mev_sum)
                belong_to_nest = BelongsTo(
                    alternative_id, set(nest.list_of_alternatives)
                )
                the_term = ConditionalTermTuple(condition=belong_to_nest, term=mev_term)

                list_of_terms.append(the_term)
            dict_of_mev_terms[i] = ConditionalSum(list_of_terms)

        corrected_utilities = {
            key: util - Variable(f"{LOG_PROBA_COL}_{key}") + dict_of_mev_terms[key]
            for key, util in self.utilities.items()
        }
        return loglogit(corrected_utilities, None, 0)

    def get_cross_nested_logit(self) -> Expression:
        """Returns the expression for the log likelihood of the nested logit model"""
        nests = self.context.cnl_nests

        # We  compute the MEV partial sum for each nest
        dict_of_mev_sums = {}

        for nest in nests:
            mu_param = nest.nest_param
            # We first build the MEV term using a sample of
            # alternatives. To build this term, we iterate from 1, as
            # we ignore the chosen alternative.
            list_of_terms = []
            for i, utility in self.mev_utilities.items():
                alpha_name = f'{self.mev_prefix}{CNL_PREFIX}{nest.name}_{i}'
                alpha = Variable(alpha_name)
                weight = Variable(f"{self.mev_prefix}{MEV_WEIGHT}_{i}")
                the_term = ConditionalTermTuple(
                    condition=alpha != 0.0,
                    term=weight * alpha**mu_param * exp(mu_param * utility),
                )
                list_of_terms.append(the_term)

            dict_of_mev_sums[nest.name] = ConditionalSum(list_of_terms)

        # We now add all relevant MEV terms to the utilities
        dict_of_mev_terms = {}
        for i, the_utility in self.utilities.items():
            list_of_terms = []
            for nest in nests:
                # Note that we need an alpha for each alternative in the main sample.
                alpha = Variable(f"{CNL_PREFIX}{nest.name}_{i}")
                mu_param = nest.nest_param
                mev_sum = dict_of_mev_sums[nest.name] ** ((1.0 / mu_param) - 1.0)
                mev_term = alpha**mu_param * exp((mu_param - 1) * the_utility) * mev_sum
                the_term = ConditionalTermTuple(condition=alpha != 0.0, term=mev_term)
                list_of_terms.append(the_term)
            dict_of_mev_terms[i] = logzero(ConditionalSum(list_of_terms))

        corrected_utilities = {
            key: util - Variable(f"{LOG_PROBA_COL}_{key}") + dict_of_mev_terms[key]
            for key, util in self.utilities.items()
        }
        return loglogit(corrected_utilities, None, 0)



"""
Generates a LaTeX output

Michel Bierlaire
Mon Sep 30 17:50:30 2024
"""

import logging
import os
from collections import Counter
from datetime import datetime

import numpy as np
from biogeme.version import get_latex, get_version, versionDate

from .estimation_results import EstimateVarianceCovariance, EstimationResults
from ..exceptions import BiogemeError
from ..parameters import Parameters
from ..tools.ellipse import Ellipse

logger = logging.getLogger(__name__)

LATEX_FILE_HEADER = r"""
\documentclass{article}
\begin{document}
"""

LATEX_FILE_FOOTER = r"""
\end{document}
"""

PARAMETERS_TABLE_HEADER = r"""
\begin{tabular}{rlr@{.}lr@{.}lr@{.}lr@{.}l}
          &              &   \multicolumn{2}{l}{}         & \multicolumn{2}{l}{__VARCOVAR__}  &  \multicolumn{4}{l}{}  \\
Parameter &              &   \multicolumn{2}{l}{Coeff.}   & \multicolumn{2}{l}{Asympt.}       & \multicolumn{4}{l}{}   \\
number    &  Description &   \multicolumn{2}{l}{estimate} & \multicolumn{2}{l}{std. error}    & \multicolumn{2}{l}{$t$-stat}  &  \multicolumn{2}{l}{$p$-value} \\
\hline
"""

PARAMETERS_TABLE_HEADER_ACTIVE = r"""
\begin{tabular}{rlr@{.}lr@{.}lr@{.}lr@{.}ll}
          &              &   \multicolumn{2}{l}{}         & \multicolumn{2}{l}{__VARCOVAR__}  &  \multicolumn{4}{l}{} & \\
Parameter &              &   \multicolumn{2}{l}{Coeff.}   & \multicolumn{2}{l}{Asympt.}       & \multicolumn{4}{l}{}  & \\
number    &  Description &   \multicolumn{2}{l}{estimate} & \multicolumn{2}{l}{std. error}    & \multicolumn{2}{l}{
$t$-stat}  &  \multicolumn{2}{l}{$p$-value} & \\
\hline
"""

PARAMETERS_TABLE_FOOTER = r"""
\end{tabular}
"""

COMPARAISON_PREAMBLE = r"""
\usepackage{longtable}
\usepackage{siunitx}
\sisetup{
  parse-numbers=false,      % Prevents automatic parsing (needed for parentheses & superscripts)
  detect-inline-weight=math,% Ensures proper formatting in tables
  tight-spacing=true        % Keeps spacing consistent
}
"""


def add_trailing_zero(formatted_number: str) -> str:
    """If the formatted number does not contain a period, we add at the end a period and a zero.

    :param formatted_number: number already formatted
    :return: process strings
    """
    if not formatted_number:
        return '0.0'
    if '.' in formatted_number:
        if formatted_number[-1] == '.':
            return f'{formatted_number}0'
        return formatted_number
    return f'{formatted_number}.0'


def format_with_period(value):
    """
    Formats a number using the '.3g' format while ensuring that a decimal point is always present.

    :param value: The number to format (int or float).
    :return: A string representation of the number in '.3g' format, always containing a period.

    Example:
    >>> format_with_period(1)
    '1.0'
    >>> format_with_period(1.23)
    '1.23'
    >>> format_with_period(1234)
    '1230'
    >>> format_with_period(1000000)
    '1e+06'
    >>> format_with_period(0.00456)
    '0.00456'
    """
    formatted = f'{value:.3g}'  # Use .3g format
    return formatted if '.' in formatted or 'e' in formatted else f'{formatted}.0'


def format_real_number(value: float) -> str:
    """Format a real number to be included in the LaTeX table"""

    formatted_value = f'{value:.3g}'

    # If the scientific notation has been used, we need to cancel it.
    if 'e' in formatted_value:
        left, right = formatted_value.split('e')
        return f'{add_trailing_zero(left)}e{right}'
    if 'E' in formatted_value:
        left, right = formatted_value.split('E')
        return f'{add_trailing_zero(left)}E{right}'
    return add_trailing_zero(formatted_value)


def _get_latex_header(estimation_results: EstimationResults) -> str:
    """Prepare the header for the LaTeX file, containing comments and the
    version of Biogeme.

    :return: string containing the header.
    """
    header = ''
    header += '%% This file is designed to be included into a LaTeX document\n'
    header += '%% See http://www.latex-project.org for information about LaTeX\n'
    if estimation_results.raw_estimation_results is None:
        header += 'No estimation result is available.'
        return header
    header += (
        f'%% {estimation_results.model_name} - Report from '
        f'biogeme {get_version()} '
        f'[{versionDate}]\n'
    )

    header += get_latex()
    return header


def get_latex_preamble(estimation_results: EstimationResults, file_name: str) -> str:
    """Generates the first part of the LaTeX file, with the preamble information.

    :param estimation_results: estimation results
    :param file_name: name of the LaTeX file (used only for reporting)
    :return: HTML code
    """
    now = datetime.now()
    latex = ''
    latex += f'\n% This file has automatically been generated on {now}\n'
    if estimation_results.raw_estimation_results is None:
        latex += '% No estimation result is available.'
        return latex
    latex += f'% Report file: {file_name}\n'
    latex += f'% Database name: {estimation_results.raw_estimation_results.data_name}\n'

    latex += _get_latex_header(estimation_results=estimation_results)

    if not estimation_results.algorithm_has_converged:
        latex += r'\section{Algorithm failed to converge.}\n'
        latex += (
            'It seems that the optimization algorithm did not converge. '
            'Therefore, the results below do not correspond to the maximum '
            'likelihood estimator. Check the specification of the model, '
            'or the criteria for convergence of the algorithm.'
        )
    identification_threshold = Parameters().get_value('identification_threshold')
    try:
        if np.abs(estimation_results.smallest_eigenvalue) <= identification_threshold:
            latex += r'\section{Warning: identification issue}\n'
            latex += (
                f'The second derivatives matrix is close to singularity. '
                f'The smallest eigenvalue is '
                f'{np.abs(estimation_results.smallest_eigenvalue):.3g}. This warning is '
                f'triggered when it is smaller than the parameter '
                f'\\texttt{{identification_threshold}}='
                f'{identification_threshold}.\n\n'
                f'Variables involved:'
            )
            latex += r'\begin{tabular}{l@{*}l}'
            for i, ev in enumerate(estimation_results.smallest_eigenvector):
                if np.abs(ev) > identification_threshold:
                    latex += (
                        f'{ev:.3g}' f' & ' f'{estimation_results.beta_names[i]}\\\\ \n'
                    )
            latex += r'\end{tabular}'
    except BiogemeError:
        latex += r'\section{Warning: second derivatives matrix not available}\n'
        latex += 'The second derivatives matrix has not been calculated. The statistics requiring it are not generated.'

    if estimation_results.user_notes is not None:
        # User notes
        latex += f'%% User notes: {estimation_results.user_notes}'
    return latex


def get_latex_general_statistics(estimation_results: EstimationResults) -> str:
    """Get the general statistics coded in LaTeX

    :return: LaTeX code
    """
    latex = '\n%% General statistics\n'
    latex += '\\section{General statistics}\n'
    statistics = estimation_results.get_general_statistics()
    latex += '\\begin{tabular}{ll}\n'
    for name, value in statistics.items():
        value = value.replace('_', '\\_')
        latex += f'{name} & {value} \\\\\n'
    for key, value in estimation_results.optimization_messages.items():
        if key in (
            'Relative projected gradient',
            'Relative change',
            'Relative gradient',
        ):
            latex += f'{key} & \\verb${value:.4g}$ \\\\\n'
        else:
            latex += f'{key} & \\verb${value}$ \\\\\n'
    latex += '\\end{tabular}\n'

    return latex


def get_latex_one_parameter(
    estimation_results: EstimationResults,
    parameter_index: int,
    variance_covariance_type: EstimateVarianceCovariance,
    parameter_number=None,
    parameter_name=None,
) -> str:
    """Generate the LaTeX code for one row of the table of the estimated parameters.

    :param estimation_results: estimation results.
    :param parameter_index: index of the parameter
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param parameter_number: number of the parameter to report. If None, it is the index.
    :param parameter_name: name of the parameter to report. If None, taken from estimation results.
    :return: LaTeX code for the row
    """
    if parameter_index < 0 or parameter_index >= len(estimation_results.beta_names):
        error_msg = f'Invalid parameter index {parameter_index}. Valid range: 0- {len(estimation_results.beta_names)-1}'
        raise ValueError(error_msg)
    if parameter_number is None:
        parameter_number = parameter_index
    if parameter_name is None:
        parameter_name = estimation_results.beta_names[parameter_index]

    output = f'{parameter_number}'
    output += ' & '
    output += parameter_name.replace('_', r'\_')
    output += ' & '
    # Value
    value = estimation_results.get_parameter_value_from_index(
        parameter_index=parameter_index
    )
    output += format_real_number(value).replace('.', '&')
    output += ' & '
    # std err
    std_err = estimation_results.get_parameter_std_err_from_index(
        parameter_index=parameter_index, estimate_var_covar=variance_covariance_type
    )
    output += format_real_number(std_err).replace('.', '&')
    output += ' & '
    # t-test against 0
    t_test = estimation_results.get_parameter_t_test_from_index(
        parameter_index=parameter_index,
        estimate_var_covar=variance_covariance_type,
        target=0,
    )
    output += format_real_number(t_test).replace('.', '&')
    output += ' & '
    # p-value against 0
    p_value = estimation_results.get_parameter_p_value_from_index(
        parameter_index=parameter_index,
        estimate_var_covar=variance_covariance_type,
        target=0,
    )
    output += format_real_number(p_value).replace('.', '&')
    if estimation_results.is_bound_active(
        parameter_name=estimation_results.beta_names[parameter_index]
    ):
        output += '& Active bound'
    output += r' \\ '
    return output


def rename_and_renumber(
    names: list[str],
    renumbering_parameters: dict[int, int] | None = None,
    renaming_parameters: dict[str, str] | None = None,
) -> dict[int, tuple[int, str]]:
    """Rename and renumber parameters according to instructions from the user

    :param names: list of existing names
    :param renumbering_parameters: dict mapping old numbers to new numbers.
    :param renaming_parameters: dict mapping old names to new names.
    :return: dict mapping the original parameter index with the new number and the new name.
    """

    if renumbering_parameters is None:
        renumbering_parameters = {}
    if renaming_parameters is None:
        renaming_parameters = {}
    else:
        # Verify that the renaming is well-defined.
        name_values = list(renaming_parameters.values())
        if len(name_values) != len(set(name_values)):
            warning_msg = f'The new renaming assigns the same name for multiple parameters. It may not be the desired action: {renaming_parameters}'
            logger.warning(warning_msg)

    updated_items = {}
    for old_key, name in enumerate(names):
        new_key = renumbering_parameters.get(
            old_key, old_key
        )  # Default to original key
        updated_items[old_key] = (new_key, renaming_parameters.get(name, name))

    list_of_new_keys = [val[0] for val in updated_items.values()]
    element_counts = Counter(list_of_new_keys)
    repeated_elements = {
        key: value for key, value in element_counts.items() if value > 1
    }
    if repeated_elements:
        error_msg = f'The following new indices appear more than once: {repeated_elements.keys()}'
        raise BiogemeError(error_msg)
    return updated_items


def get_latex_estimated_parameters(
    estimation_results: EstimationResults,
    variance_covariance_type: EstimateVarianceCovariance,
    renumbering_parameters: dict[int, int] | None = None,
    renaming_parameters: dict[str, str] | None = None,
) -> str:
    """Get the estimated parameters coded in LaTeX

    :param estimation_results: estimation results.
    :param variance_covariance_type: type of variance-covariance estimate to be used.
    :param renumbering_parameters: a dict that suggests new numbers for parameters
    :param renaming_parameters: a dict that suggests new names for some or all parameters.
    :return: LaTeX code
    """

    if renumbering_parameters is not None:
        # Verify that the numbering is well-defined
        number_values = list(renumbering_parameters.values())
        if len(number_values) != len(estimation_results.beta_names):
            error_msg = (
                f'The new numbering involves {len(number_values)} values while there are '
                f'{len(estimation_results.beta_names)} parameters'
            )
            raise BiogemeError(error_msg)

        if len(number_values) != len(set(number_values)):
            error_msg = f'The new numbering cannot assign the same number to two different parameters: {renumbering_parameters}'
            raise BiogemeError(error_msg)

    the_header = (
        PARAMETERS_TABLE_HEADER_ACTIVE
        if estimation_results.is_any_bound_active()
        else PARAMETERS_TABLE_HEADER
    )
    output = the_header.replace('__VARCOVAR__', str(variance_covariance_type))

    renamed_parameters: dict[int, tuple[int, str]] = rename_and_renumber(
        names=estimation_results.beta_names,
        renumbering_parameters=renumbering_parameters,
        renaming_parameters=renaming_parameters,
    )
    all_rows = {}
    for old_index, user_defined in renamed_parameters.items():
        new_index, new_name = user_defined
        the_row = (
            get_latex_one_parameter(
                estimation_results=estimation_results,
                parameter_index=old_index,
                variance_covariance_type=variance_covariance_type,
                parameter_number=new_index,
                parameter_name=new_name,
            )
            + '\n'
        )
        all_rows[new_index] = the_row
    for a_row_number in sorted(all_rows):
        output += all_rows[a_row_number]
    output += PARAMETERS_TABLE_FOOTER
    return output


def get_sign_for_p_value(p_value: float, p_thresholds: list[tuple[float, str]]) -> str:
    """

    :param p_value: p-value to be treated
    :param p_thresholds: list of tuple establishing the coding convention for the p-value threshold. Assume that the
    list is composed of pairs (t_i, sign_i), and that the p-value of a parameter is p. Among all i such that p <=
    t_i, select index k associated with the minimum t_k. Then the p-value is coded using sign_k.
    :return: sign to be used
    """
    try:
        the_p_value = float(p_value)
    except (ValueError, TypeError) as e:
        raise TypeError(f'Cannot convert {p_value} to float') from e

    try:
        candidates_thresholds = [
            (float(t), str(sign)) for t, sign in p_thresholds if the_p_value <= float(t)
        ]
    except (ValueError, TypeError) as e:
        raise TypeError(f'Non float thresholds: {p_thresholds}') from e

    if not candidates_thresholds:
        return ''

    _, selected_sign = min(candidates_thresholds, key=lambda x: x[0])
    return selected_sign


def compare_parameters(
    estimation_results: dict[str, EstimationResults],
    p_thresholds: list[tuple[float, str]] | None = None,
    renumbering_parameters: dict[int, int] | None = None,
    renaming_parameters: dict[str, str] | None = None,
) -> str:
    """

    :param estimation_results: dict mapping model names with estimation results
    :param p_thresholds: list of tuple establishing the coding convention for the p-value threshold. Assume that the
    list is composed of pairs (t_i, sign_i), and that the p-value of a parameter is p. Among all i such that p <=
    t_i, select index k associated with the minimum t_k. Then the p-value is coded using sign_k.
    :param renumbering_parameters: a dict that suggests new numbers for parameters
    :param renaming_parameters: a dict that suggests new names for some or all parameters.
    :return: the LaTeX code of a tabular object compile the estimation results.
    """
    if p_thresholds is None:
        p_thresholds = [
            (0.01, r'\textsuperscript{***}'),
            (0.05, r'\textsuperscript{**}'),
            (0.1, r'\textsuperscript{*}'),
        ]

    latex_code = COMPARAISON_PREAMBLE
    # Opening the longtable environment
    number_of_columns = len(estimation_results)
    s_columns = 'S' * number_of_columns
    latex_code += r'\begin{longtable}{rl' + s_columns + r'}'
    latex_code += '\n'
    # First row
    latex_code += '& & '
    latex_code += ' & '.join(
        [
            fr'\multicolumn{{1}}{{c}}{{{model_name}}}'
            for model_name in estimation_results.keys()
        ]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # Second row
    latex_code += ' & Parameter name & '
    latex_code += ' & '.join(
        [r' \multicolumn{1}{c}{Coef./(SE)}' for _ in estimation_results.keys()]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # Separation line
    latex_code += r'\hline'
    latex_code += '\n'
    # Generate the list of all parameters
    all_beta_names: set[str] = set()
    for estimation_result in estimation_results.values():
        all_beta_names.update(estimation_result.beta_names)

    renamed_parameters: dict[int, tuple[int, str]] = rename_and_renumber(
        names=sorted(all_beta_names),
        renumbering_parameters=renumbering_parameters,
        renaming_parameters=renaming_parameters,
    )

    rows_1 = []
    rows_2 = []
    for new_id, name in renamed_parameters.values():
        row_1 = ''
        row_2 = ''
        row_1 += f'{new_id} &'
        row_2 += ' & '
        row_1 += name.replace('_', r'\_')
        for estimation_result in estimation_results.values():
            if name in estimation_result.beta_names:
                value = estimation_result.get_parameter_value(parameter_name=name)
                std_err = estimation_result.get_parameter_std_err(parameter_name=name)
                p_value = estimation_result.get_parameter_p_value(parameter_name=name)
                sign = get_sign_for_p_value(p_value, p_thresholds)
                formatted_value = format_with_period(value)
                row_1 += f'& {formatted_value}{sign} '
                row_2 += f'&({std_err:.3g})'
            else:
                row_1 += ' & '
                row_2 += ' & '
        row_1 += r' \\'
        row_2 += r' \\'
        row_1 += '\n'
        row_2 += '\n'
        rows_1.append(row_1)
        rows_2.append(row_2)

    for row_1, row_2 in zip(rows_1, rows_2):
        latex_code += row_1
        latex_code += row_2

    # Separation line
    latex_code += r'\hline'
    latex_code += '\n'

    # Observations
    latex_code += r'\multicolumn{2}{l}{Number of observations} &'
    latex_code += ' & '.join(
        [
            f'{estimation_results.sample_size}'
            for estimation_results in estimation_results.values()
        ]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # Parameters
    latex_code += r'\multicolumn{2}{l}{Number of parameters} &'
    latex_code += ' & '.join(
        [
            f'{estimation_results.number_of_free_parameters}'
            for estimation_results in estimation_results.values()
        ]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # AIC
    latex_code += r'\multicolumn{2}{l}{Akaike Information Criterion} &'
    latex_code += ' & '.join(
        [
            f'{estimation_results.akaike_information_criterion:.1f}'
            for estimation_results in estimation_results.values()
        ]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # BIC
    latex_code += r'\multicolumn{2}{l}{Bayesian Information Criterion} &'
    latex_code += ' & '.join(
        [
            f'{estimation_results.bayesian_information_criterion:.1f}'
            for estimation_results in estimation_results.values()
        ]
    )
    latex_code += r' \\'
    latex_code += '\n'

    # Separation line
    latex_code += r'\hline'
    latex_code += '\n'

    latex_code += r'\multicolumn{4}{l}{\footnotesize Standard errors: '
    footnote = [f'{mark}: $p < {threshold}$' for (threshold, mark) in p_thresholds]
    latex_code += ', '.join(footnote)
    latex_code += r'}'
    latex_code += '\n'

    # Closing the longtable environment
    latex_code += r'\end{longtable}'
    latex_code += '\n'
    return latex_code


def draw_confidence_ellipse(
    ellipse: Ellipse,
    first_reporting_name: str,
    second_reporting_name: str,
) -> str:
    """Provides a Tikz picture of the confidence ellipsis for two parameters
    :param ellipse: ellipse to be drawn
    :param first_reporting_name: name to use in the reporting for the first parameter
    :param second_reporting_name: name to use in the reporting for the second parameter
    """
    # Define axis ranges
    xmin, xmax = (
        ellipse.center_x - 1.5 * ellipse.axis_one,
        ellipse.center_x + 1.5 * ellipse.axis_one,
    )
    ymin, ymax = (
        ellipse.center_y - 1.5 * ellipse.axis_two,
        ellipse.center_y + 1.5 * ellipse.axis_two,
    )

    # Generate TikZ code
    tikz_code = f"""
    \\begin{{tikzpicture}}
        \\begin{{axis}}[
            width=13cm, height=9cm,
            grid=major,
            axis x line=middle,
            axis y line=middle,
            xlabel={{{first_reporting_name}}},
            ylabel={{{second_reporting_name}}},
            xmin={xmin}, xmax={xmax},
            ymin={ymin}, ymax={ymax},
            axis line style={{thick}},
            ticks=both
        ]
            % Ellipse parameters
            \\pgfmathsetmacro{{\\centerx}}{{{ellipse.center_x:.4f}}}
            \\pgfmathsetmacro{{\\centery}}{{{ellipse.center_y:.4f}}}
            \\pgfmathsetmacro{{\\cosphi}}{{{ellipse.cos_phi:.6f}}}
            \\pgfmathsetmacro{{\\sinphi}}{{{ellipse.sin_phi:.6f}}}
            \\pgfmathsetmacro{{\\axisone}}{{{ellipse.axis_one:.6f}}}
            \\pgfmathsetmacro{{\\axistwo}}{{{ellipse.axis_two:.6f}}}

            % Center of the ellipse
            \\coordinate (center) at (\\centerx, \\centery);
            \\node at (center) {{$\\bullet$}};

            % Plot ellipse
            \\addplot[domain=0:360, samples=100, smooth, thick] (
                {{\\centerx + \\axisone * cos(x) * \\cosphi - \\axistwo * sin(x) * \\sinphi}},
                {{\\centery + \\axisone * cos(x) * \\sinphi + \\axistwo * sin(x) * \\cosphi}}
            );
        \\end{{axis}}
    \\end{{tikzpicture}}
    """
    return tikz_code


def generate_latex_file_content(
    estimation_results: EstimationResults,
    filename: str,
    variance_covariance_type: EstimateVarianceCovariance,
    include_begin_document=False,
) -> str:
    """
    Generate the full content of a LaTeX document summarizing the estimation results,
    and prepare it for export to a file.

    This function assembles a complete LaTeX document including the document header,
    general statistics, and the estimated parameter table. It is intended to be used
    to produce a full report from Biogeme estimation results.

    :param estimation_results: The estimation results object containing the model outputs.
    :param filename: The name of the LaTeX file, used only for documentation within the file.
    :param include_begin_document: if True, the LaTeX file can be directly compiled. If False, it must be included
        in another document.
    :param variance_covariance_type: select which type of variance-covariance matrix is used to generate the
        statistics.
    :return: None. The LaTeX content is generated as a string but not written to a file.
    """
    latex = ''
    if include_begin_document:
        latex += LATEX_FILE_HEADER
    latex += get_latex_preamble(
        estimation_results=estimation_results, file_name=filename
    )
    latex += get_latex_general_statistics(estimation_results=estimation_results)
    latex += get_latex_estimated_parameters(
        estimation_results=estimation_results,
        variance_covariance_type=variance_covariance_type,
    )
    if include_begin_document:
        latex += LATEX_FILE_FOOTER
    return latex


def generate_latex_file(
    estimation_results: EstimationResults,
    filename: str,
    include_begin_document=False,
    overwrite=False,
    variance_covariance_type: EstimateVarianceCovariance | None = None,
) -> None:
    """
    Generate and save a LaTeX document that summarizes the model estimation results.

    This function calls `generate_latex_file_content` to build the full LaTeX document,
    and writes it to a file. If the file already exists and `overwrite` is False,
    a `FileExistsError` is raised.

    :param estimation_results: Estimation results to be documented.
    :param include_begin_document: if True, the LaTeX file can be directly compiled. If False, it must be included
        in another document.
    :param filename: Path to the LaTeX file to be created.
    :param overwrite: If True, overwrite the file if it already exists. Defaults to False.
    :param variance_covariance_type: select which type of variance-covariance matrix is used to generate the
        statistics. If None, the bootstrap one is used if available. If not available, the robust one.
    """
    if variance_covariance_type is None:
        variance_covariance_type = (
            estimation_results.get_default_variance_covariance_matrix()
        )
    if (
        variance_covariance_type == EstimateVarianceCovariance.BOOTSTRAP
        and estimation_results.bootstrap_time is None
    ):
        logger.warning(
            f'No bootstrap data is available. The robust variance-covariance matrix is used instead.'
        )
        variance_covariance_type = EstimateVarianceCovariance.ROBUST

    if not overwrite and os.path.exists(filename):
        raise FileExistsError(f"The file '{filename}' already exists.")

    with open(filename, 'w') as file:
        content = generate_latex_file_content(
            estimation_results=estimation_results,
            filename=filename,
            include_begin_document=include_begin_document,
            variance_covariance_type=variance_covariance_type,
        )
        print(content, file=file)
    logger.info(f'File {filename} has been generated.')



"""Tests for audit functions

Michel Bierlaire
Fri Mar 28 17:35:27 2025
"""

import unittest

from biogeme.expressions import (
    BelongsTo,
    Beta,
    Draws,
    IntegrateNormal,
    LogLogit,
    MonteCarlo,
    RandomVariable,
    Variable,
)
from biogeme.expressions.audit import audit_expression


class TestAuditFunctions(unittest.TestCase):
    def test_montecarlo_with_draws(self):
        expr = MonteCarlo(Draws('eps', 'NORMAL'))
        errors, warnings = audit_expression(expr)
        self.assertEqual(errors, [])
        self.assertEqual(warnings, [])

    def test_montecarlo_without_draws(self):
        expr = MonteCarlo(Beta('beta_1', 1.0, None, None, 0))
        errors, warnings = audit_expression(expr)
        self.assertTrue(any('does not contain any Draws' in e for e in errors))

    def test_montecarlo_nested(self):
        expr = MonteCarlo(MonteCarlo(Draws('eps', 'NORMAL')))
        errors, warnings = audit_expression(expr)
        self.assertTrue(any('cannot contain another MonteCarlo' in e for e in errors))

    def test_integrate_with_random_variable(self):
        expr = IntegrateNormal(RandomVariable('x'), 'x')
        errors, warnings = audit_expression(expr)
        self.assertEqual(errors, [])
        self.assertEqual(warnings, [])

    def test_integrate_without_random_variable(self):
        expr = IntegrateNormal(Beta('beta_1', 1.0, None, None, 0), 'x')
        errors, warnings = audit_expression(expr)
        self.assertFalse(errors)
        self.assertTrue(
            any('does not contain any RandomVariable' in w for w in warnings)
        )

    def test_belongsto_with_non_integer(self):
        expr = BelongsTo(Variable('x'), {1.0, 2.5, 3})
        errors, warnings = audit_expression(expr)
        self.assertTrue(any('not integer' in w for w in warnings))

    def test_loglogit_with_mismatched_keys(self):
        util = {1: Beta('b1', 1.0, None, None, 0)}
        av = {2: Beta('a1', 1.0, None, None, 0)}
        expr = LogLogit(util, av, 1)
        errors, warnings = audit_expression(expr)
        self.assertTrue(any('Incompatible list of alternatives' in e for e in errors))

    def test_loglogit_with_matching_keys(self):
        util = {1: Beta('b1', 1.0, None, None, 0)}
        av = {1: Beta('a1', 1.0, None, None, 0)}
        expr = LogLogit(util, av, 1)
        errors, warnings = audit_expression(expr)
        self.assertEqual(errors, [])
        self.assertEqual(warnings, [])


if __name__ == '__main__':
    unittest.main()



"""
PanelStructure: Handles organization and indexing of panel data,
where observations are grouped by individuals.

Michel Bierlaire
Wed Mar 26 19:33:13 2025
"""

import logging

import pandas as pd

from .container import Database

RELEVANT_PREFIX = 'relevant_'

logger = logging.getLogger(__name__)


def observation_suffix(index: int) -> str:
    """Return a zero-padded suffix for observation index (1-based)."""
    return f"__panel__{index + 1:02d}"


class PanelDatabase:
    """Encapsulates a pandas DataFrame for Biogeme, in panel format,"""

    def __init__(self, database: Database, panel_column: str):
        """
        Constructor

        :param database: original database
        :param panel_column: name of the column with the identifier of the individuals.
        :raises BiogemeError: if the dataframe is empty
        """
        self.original_database = database
        # Entries for panel data
        self.panel_column: str = panel_column

    def __str__(self) -> str:
        return f'biogeme panel database {self.database.name}'

    def flatten_database(self, missing_data: float) -> tuple[pd.DataFrame, int]:
        """
        Flatten a long-format dataframe into a wide-format dataframe where each row represents one individual
        and columns represent multiple observations.

        :param missing_data: value to use if data is missing.
        :return: a tuple containing two things:
            - A wide-format DataFrame where each row corresponds to one individual.
            For each variable column in the original DataFrame (excluding `grouping_column`), the output contains
            multiple columns named `columnname_XX`, where `XX` is the zero-padded observation index (starting at 01).
            Additionally, for each observation index, a `relevant_XX` column indicates whether the observation
            is relevant (1) or padded with a missing value (0).

            - The size of the largest group.

        """
        dataframe = self.original_database.dataframe
        logger.info(f'Flattening database [{dataframe.shape}].')

        grouping_column = self.panel_column
        if dataframe.empty:
            return dataframe.copy(), 0

        # Find non-ID columns
        value_columns = [col for col in dataframe.columns if col != grouping_column]

        # Identify columns with constant values within each group
        constant_columns = []
        for col in value_columns:
            if dataframe.groupby(grouping_column)[col].nunique().max() == 1:
                constant_columns.append(col)

        # Determine max number of observations per individual
        largest_group = dataframe.groupby(grouping_column).size().max()

        # Prepare list to collect rows
        flattened_rows = []

        for individual_id, group in dataframe.groupby(grouping_column):
            group = group.reset_index(drop=True)
            obs_count = len(group)
            row = {grouping_column: individual_id}
            for col in constant_columns:
                row[col] = group.at[0, col]

            for obs_index in range(largest_group):
                suffix = observation_suffix(obs_index)
                row[f"{RELEVANT_PREFIX}{suffix}"] = 1 if obs_index < obs_count else 0
                is_valid = obs_index < obs_count
                for col in value_columns:
                    value = group.at[obs_index, col] if is_valid else missing_data
                    row[f"{col}{suffix}"] = value
            flattened_rows.append(row)

        flat_dataframe = pd.DataFrame(flattened_rows)
        logger.info(f'Database flattened [{flat_dataframe.shape}]')

        return flat_dataframe, largest_group
